#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""inspire.py
Collection of functions useful for the INSPIRE challenge.

"""
from __future__ import absolute_import, division, print_function, unicode_literals
import copy
import gzip

import gzip
import json
import logging
import os
import pprint as pp
import collections
import re
import shlex
import urllib2
import io
import itertools
import requests
import requests.exceptions
import zipfile
import StringIO
import progressbar as pb
from scipy.io import wavfile
from contextlib import closing

from . import common
from . import edit_distance
import time

UTF8_NORMALIZATION = 'NFD'

BASE_URL = 'http://46.226.110.12:5000'
BASE_URL = 'http://localhost:5000'


def _load_zip_wav(zfile, offset=0, count=None):
    """Load a wav file into an array from frame start to fram end

    :param zfile: ZipExtFile file-like object from where to load the audio
    :param offset: First sample to load
    :param count: Maximum number of samples to load
    :return: The audio samples in a numpy array of floats
    """

    buf = StringIO.StringIO(zfile.read())

    sample_rate, audio = wavfile.read(buf)
    audio = audio[offset:]

    if count:
        audio = audio[:count]

    return sample_rate, audio


def _get_url(url):
    r = requests.get(url)

    if r.status_code != 200:
        logging.error('Could not get: \n'
                      '{}\n'
                      'Please check your internet connection.\n'
                      'If problem persists inform the challenge organizers.'.format(url))
        return '{}'

    return r.content


def _download_url(url, filename=None):
    with closing(requests.get(url, stream=True)) as r:
        if r.status_code != 200:
            logging.error('Could not download: \n'
                          '{}\n'
                          'Please check your internet connection.\n'
                          'If problem persists inform the challenge organizers.'.format(url))
            return None

        reported_filename = os.path.basename(url)
        if r.headers.get('Content-Disposition'):
            reported_filename = shlex.split(re.findall(r'filename=(\S+)',
                                                       r.headers['Content-Disposition'])[0])[0]

        filename = filename or reported_filename

        if os.path.isfile(filename):
            return filename

        with open(filename, 'wb') as f:
            file_size = r.headers.get('content-length')

            if file_size is None:  # no content length header
                f.write(r.content)
            else:
                file_size_dl = 0
                file_size = int(file_size)

                widgets = ['Downloading: ', pb.Percentage(), ' ', pb.Bar(),
                           ' ', pb.ETA(), ' ', pb.FileTransferSpeed()]
                pbar = pb.ProgressBar(widgets=widgets, maxval=file_size).start()
                chunk_size = 10240
                for data in r.iter_content(chunk_size=chunk_size):
                    file_size_dl = min(file_size_dl + chunk_size, file_size)
                    f.write(data)

                    pbar.update(file_size_dl)

                pbar.finish()

            print('Finished downloading {}'.format(filename))

    return filename


class MyPrettyPrinter(pp.PrettyPrinter):
    _escape = dict((q, dict(((c, unicode(repr(chr(c)))[1:-1])
                             for c in range(32) + [ord('\\')] +
                             range(128, 161)),
                            **{ord(q): u'\\' + q}))
                   for q in ["'", '"'])

    def format(self, obj, context, maxlevels, level):
        if type(obj) is unicode:
            q = "'" if "'" not in obj or '"' in obj \
                else '"'
            return ("u" + q + obj.translate(self._escape[q]) +
                    q, True, False)
        return pp.PrettyPrinter.format(
            self, obj, context, maxlevels, level)


def pprint(obj, **kwargs):
    printer = MyPrettyPrinter(**kwargs)
    printer.pprint(obj)


class EvaluationSetting(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __repr__(self):
        return json.dumps(self, indent=2)

    def download_dataset(self, filename=None):
        return _download_url('{}/download/dataset/{}'.format(BASE_URL, self['id']), filename=filename)

    def download_lexicon(self, filename=None):
        return _download_url('{}/download/lexicon/{}'.format(BASE_URL, self['id']), filename=filename)


class Dataset(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __repr__(self):
        return json.dumps(self, indent=2)


class Lexicon(dict):
    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, key):
        return dict.__getitem__(self, key.upper())

    def __setitem__(self, key, value):
        return dict.__setitem__(self, key.upper(), value)

    def __contains__(self, key):
        return dict.__contains__(self, key.upper())

    def __repr__(self):
        return json.dumps(self, indent=2)


class Submission(dict):
    def __init__(self, email=None, evaluation_setting=None, description=None, metadata={}):
        self['metadata'] = {'email': email,
                            'evaluation_setting': evaluation_setting,
                            'description': description}

        self['metadata'].update(metadata)

        if self['metadata']['evaluation_setting'] is None:
            raise ValueError('Must set the evaluation_setting, when constructing a Submission.')

        self['tokens'] = {}

    def where_task(self, token_id, confusion_probability):
        """Provide the prediction of the where task.

        This function is used to predict the probability of a given pronunciation being reported for a given token.

        :param token_id: The token for which the prediction is being provided
        :param confusion_probability: The list or array of confusion probabilities at each index
        :param phonemes_probability: The probability of the phoneme or phoneme sequence
        """

        self['tokens'].setdefault(token_id, {})['where'] = list(confusion_probability)

    def what_task(self, token_id, index, phonemes, phonemes_probability, warn=True):
        """Provide the prediction of the what task.

        This function is used to predict the probability of a given phoneme being reported at a given index
        for a given token.

        :param token_id: The token for which the prediction is provided
        :param index: The index of the token for which the prediction is provided
        :param phonemes: The phoneme or phoneme sequence for which the prediction is being made
        (as a space separated string)
        :param phonemes_probability: The probability of the phoneme or phoneme sequence
        :param warn: Set to False in order to avoid warnings about 0 or 1 probabilities
        """

        if not 0. < phonemes_probability < 1. and warn:
            logging.warning('Setting a probability of [{}] to phonemes [{}] for token [{}].\n '
                            'Using probabilities of 0.0 or 1.0 '
                            'may lead to likelihoods of -Infinity'.format(phonemes_probability,
                                                                          phonemes,
                                                                          token_id))

        self['tokens'].setdefault(token_id, {}) \
            .setdefault('what', {}) \
            .setdefault(str(index), {})[phonemes] = phonemes_probability

    def full_task(self, token_id, pronunciation, pronunciation_probability, warn=True):
        """Provide the prediction of the full task.

        This function is used to predict the probability of a given pronunciation being reported for a given token.

        :param token_id: The token for which the prediction is provided
        :param pronunciation: The pronunciation for which the prediction is being made (as a list of strings
        or space separated string)
        :param pronunciation_probability: The probability of the pronunciation for the given token
        :param warn: Set to False in order to avoid warnings about 0 or 1 probabilities
        """

        if not 0. < pronunciation_probability < 1. and warn:
            logging.warning('Setting a probability of [{}] to pronunciation [{}] for token [{}].\n '
                            'Using probabilities of 0.0 or 1.0 '
                            'may lead to likelihoods of -Infinity'.format(pronunciation_probability,
                                                                          pronunciation,
                                                                          token_id))

        key = pronunciation
        if isinstance(key, list):
            if not all([isinstance(phoneme, basestring) for phoneme in key]):
                raise ValueError('The pronunciation must be of type string (a sequence of space separated phonemes) '
                                 'or of type list (containing phonemes of type strings).'
                                 'User supplied: {}'.format(key))

            key = ' '.join(pronunciation)

        self['tokens'].setdefault(token_id, {}) \
            .setdefault('full', {})[key] = pronunciation_probability

    def dumps(self):
        bytes = io.BytesIO()
        self.dump(bytes)
        return bytes.getvalue()

    def dump(self, fileobj):
        with gzip.GzipFile(fileobj=fileobj, mode='w') as z:
            metadata = copy.copy(self['metadata'])
            metadata['token_count'] = len(self['tokens'])

            z.write(json.dumps(metadata))
            z.write('\n')

            for token_id, token in self['tokens'].items():
                z.write(json.dumps((token_id, token), sort_keys=True))
                z.write('\n')

    def save(self, filename):
        """Save the submission into a file.

        :param filename: where to save the submission
        """
        with open(filename, 'w') as f:
            self.dump(f)

        return

    # TODO: use a limit when ready http://bugs.python.org/issue15955

    @staticmethod
    def open(filename):
        """Open the submission from a file.

        :param filename: where to load the submission from
        """
        with open(filename, 'r') as f:
            return self.load(f)

    @staticmethod
    def open_metadata(filename):
        """Open the submission from a file.

        :param filename: where to load the submission from
        """
        with open(filename, 'r') as f:
            return self.load_metadata(f)

    @staticmethod
    def open_tokens(filename):
        """Open the submission from a file.

        :param filename: where to load the submission from
        """
        with open(filename, 'r') as f:
            return self.load_tokens(f)

    @staticmethod
    def loads(data):
        fileobj = io.BytesIO(data)
        return Submission.load(fileobj)

    @staticmethod
    def load(fileobj):
        """Load the submission from a file-like object

        :param fileobj: File-like object
        :return: the loaded submission
        """
        with gzip.GzipFile(fileobj=fileobj, mode='r') as z:
            submission = Submission(metadata=json.loads(z.readline()))

            for line in z:
                token_id, token = json.loads(line)
                submission['tokens'][token_id] = token

        return submission

    @staticmethod
    def loads_metadata(data):
        fileobj = io.BytesIO(data)
        return Submission.load_metadata(fileobj)

    @staticmethod
    def load_metadata(fileobj):
        """Load the submission from a file.

        :param filename: where to load the submission from
        """
        with gzip.GzipFile(fileobj=fileobj, mode='r') as z:
            return json.loads(z.readline())

    @staticmethod
    def loads_tokens(data):
        fileobj = io.BytesIO(data)
        return Submission.load_tokens(fileobj)

    @staticmethod
    def load_tokens(fileobj):
        with gzip.GzipFile(fileobj=fileobj, mode='r') as z:
            z.readline()  # skip the metadata

            for line in z:
                token = json.loads(line)

                yield token

    def submit(self, password=''):
        """Submits the participation to the web site.

        The passwords is sent as plain text.

        :return: the evaluation results.
        """

        url = '{}/api/submit'.format(BASE_URL)
        try:
            r = requests.post(url,
                              data=self.dumps(),
                              headers={'content-type': 'application/json'},
                              auth=(self['metadata']['email'], password))

            response = r.json()

        except requests.exceptions.HTTPError as e:
            logging.error('Error while submitting the participation. {}'.format(e))
            return Job()

        if 'error' in response:
            logging.error('Error while processing the participation. {}'.format(response['error']))
            return Job()

        return Job(response)

    def evaluate(self, password=''):
        """Evaluates the development set.

        The passwords is sent as plain text.

        :return: the evaluation results.
        """

        # Make a copy only keeping the development set
        dev_submission = self
        if self['metadata'].get('evaluation_setting', {}).get('development_set', None):
            dev_submission = copy.deepcopy(self)
            dev_submission['tokens'] = {token_id: token for token_id, token in self['tokens'].items()
                                        if token_id in self['metadata']['evaluation_setting']['development_set']}

        url = '{}/api/evaluate'.format(BASE_URL)
        try:
            r = requests.post(url,
                              data=dev_submission.dumps(),
                              headers={'content-type': 'application/json'},
                              auth=(dev_submission['metadata']['email'], password))

            response = r.json()

        except requests.exceptions.HTTPError as e:
            logging.error('Error while submitting the participation. {}'.format(e))
            return Job()

        if 'error' in response:
            logging.error('Error while processing the participation. {}'.format(response['error']))
            return Job()

        return Job(response)

    def __repr__(self):
        return json.dumps(self, indent=2)


class Job(dict):
    _ready_states = {'SUCCESS', 'FAILURE'}

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __repr__(self):
        return json.dumps(self, indent=2)

    def status(self):
        if 'task_id' not in self:
            logging.error('This job has not succeed.')
            return {}

        url = '{}/api/task/{}/status'.format(BASE_URL, self['task_id'])
        r = requests.get(url)

        return r.json()

    def wait(self):
        widgets = ['Processing: ', pb.Percentage(), ' ', pb.Bar(), ' ', pb.ETA()]

        pbar = pb.ProgressBar(widgets=widgets, maxval=100).start()

        while True:
            task_status = self.status()['task']

            if task_status['status'] == 'PROGRESS':
                if 'count' in task_status['result'] and 'total' in task_status['result']:
                    progress = int(float(task_status['result']['count']) / task_status['result']['total'] * 100)
                else:
                    progress = None

                pbar.update(progress)

            elif task_status['status'] in self._ready_states:
                pbar.finish()
                break

            else:
                pbar.update()

            time.sleep(1.0)

        return

    def result(self):
        if 'task_id' not in self:
            logging.error('This job has not succeed.')
            return {}

        url = '{}/api/task/{}/result'.format(BASE_URL, self['task_id'])
        r = requests.get(url)

        data = r.json()

        return data['task']['result']

    def ready(self):
        return self.status()['task']['status'] in self._ready_states


def download_dataset_audio(filename='dataset_audio.zip'):
    return _download_url('{}/static/audio.zip'.format(BASE_URL), filename=filename)


def get_token_audio(token_id, dataset_audio_filename, dataset):
    token = dataset['tokens'][token_id]

    with zipfile.ZipFile(dataset_audio_filename) as zf:
        signal_name = 'confusionWavs/wavs/{}'.format(token['signal_wav'])

        with zf.open(signal_name) as f:
            sample_rate, signal_audio = _load_zip_wav(f)

    speech_audio = signal_audio[:, 0]
    noise_audio = signal_audio[:, 1]

    return sample_rate, speech_audio, noise_audio


def get_evaluation_setting(setting_id=None):
    evaluation_setting_url = '{}/api/evaluation_setting'.format(BASE_URL)
    if setting_id is not None:
        evaluation_setting_url += '/{}'.format(setting_id)

    return EvaluationSetting(json.loads(_get_url(evaluation_setting_url)))


def load_wordlist(wordlist_filename):
    return set(common.parse_wordlist(wordlist_filename))


def load_lexicon(lexicon_filename):
    return Lexicon({k: [pron.split(' ') for pron in v]
                    for k, v in common.parse_dictionary(lexicon_filename).items()})


def loads_lexicon(lexicon_file):
    return Lexicon({k: [pron.split(' ') for pron in v]
                    for k, v in common.parses_dictionary(lexicon_file).items()})


def load_dataset(dataset_filename):
    with open(dataset_filename) as dataset_file:
        return loads_dataset(dataset_file)

    return {}


def loads_dataset(dataset_file):
    return Dataset(json.load(dataset_file))


def _combine_dicts(*args):
    all_dict = []
    for arg in args:
        all_dict += list(arg.items())

    return dict(all_dict)


def get_edit_scripts(pron_a, pron_b, edit_costs=(1.0, 1.0, 1.0)):
    """Get the edit scripts to transform between two given pronunciations.

    :param pron_a: Source pronunciation as list of strings, each string corresponding to a phoneme
    :param pron_b: Target pronunciation as list of strings, each string corresponding to a phoneme
    :param edit_costs: Costs of insert, replace and delete respectively
    :return: List of edit scripts.  Each edit script is represented as a list of operations,
                where each operation is a dictionary.
    """
    op_costs = {'insert': lambda x: edit_costs[0],
                'match': lambda x, y: 0 if x == y else edit_costs[1],
                'delete': lambda x: edit_costs[2]}

    distance, scripts, costs, ops = edit_distance.best_transforms(pron_a, pron_b, op_costs=op_costs)

    return [full_edit_script(script.to_primitive()) for script in scripts]


def full_edit_script(script):
    previous_index = None
    last_index = -1

    new_script = []

    for op in script:
        if op['index'] != previous_index:
            new_op = copy.copy(op)

            # Convert None to ''
            new_op['from_symbol'] = new_op['from_symbol'] or ''
            new_op['to_symbol'] = new_op['to_symbol'] or ''

            previous_index = op['index']

            # Fill gaps
            while last_index < op['index']-1:
                new_script.append({'index': last_index + 1,
                                   'from_symbol': '',
                                   'to_symbol': '',
                                   'op_code': 'noninsert'})

                last_index = new_script[-1]['index']

            new_script.append(new_op)
            last_index = new_script[-1]['index']

        else:
            # Merge consecutive inserted symbols
            if new_script[-1]['to_symbol']:
                new_script[-1]['to_symbol'] += ' {}'.format(op['to_symbol'])

            # Merge consecutive deleted symbols
            # (this doesn't happen since they have different indices)
            if new_script[-1]['from_symbol']:
                new_script[-1]['from_symbol'] += ' {}'.format(op['from_symbol'])

    if last_index % 2 == 1:
        # Missing the last gap
        new_script.append({'index': last_index + 1,
                           'from_symbol': '',
                           'to_symbol': '',
                           'op_code': 'noninsert'})

    return new_script


def print_edit_script(edit_script):
    """Print an edit script to the terminal.

    :param edit_script: The edit script as a list of operations, where each operation is a dictionary.
    """
    print('{}\n{}'.format(*edit_script_to_strings(edit_script)))


def edit_script_to_strings(edit_script, use_colors=True):
    """Convert an edit script to a pair of strings representing the operation in a human readable way.

    :param edit_script: The edit script as a list of operations, where each operation is a dictionary.
    :param use_colors: Boolean indicating whether to use terminal color codes to color the output.
    :return: Tuple with text corresponding to the first pronunciation and the text of the second one.
    """
    colors = collections.defaultdict(str)

    if use_colors:
        colors['red'] = '\x1b[31m'
        colors['normal'] = '\x1b[m'
        colors['green'] = '\x1b[32m'
        colors['on_red'] = '\x1b[41m'

    src_txt = ''
    dst_txt = ''
    for op in edit_script:
        if op['op_code'] == 'match':
            width = max(len(op['from_symbol']), len(op['to_symbol']))
            if op['from_symbol'] == op['to_symbol']:
                src_txt += '{green}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                            op,
                                                                                            {'width': width}))
                dst_txt += '{green}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                          op,
                                                                                          {'width': width}))
            else:
                src_txt += '{red}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                          op,
                                                                                          {'width': width}))
                dst_txt += '{red}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                        op,
                                                                                        {'width': width}))

        elif op['op_code'] == 'insert':
            space = ' '*len(op['to_symbol'])
            src_txt += '{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors,  op))
            dst_txt += '{red}{to_symbol}{normal}'.format(**_combine_dicts(colors, op))

        elif op['op_code'] == 'delete':
            space = ' '*len(op['from_symbol'])
            src_txt += '{red}{from_symbol}{normal}'.format(**_combine_dicts(colors, op))
            dst_txt += '{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors, op))

        elif op['op_code'] == 'noninsert':
            continue

        src_txt += ' '
        dst_txt += ' '

    return src_txt, dst_txt




#!/usr/bin/env python

import os
import logging
import subprocess
import sys
import tempfile


def temp_path(p):
    fd, tmp_path = tempfile.mkstemp(p)
    return tmp_path


def sh_command(cmd):
    subprocess.check_output(cmd, shell=True)


def htk_command(cmd):
    ret = subprocess.call(cmd, shell=True)

    if ret:
        logging.error("HTK command {} exited with error code {}".format(cmd, ret))
        sys.exit(1)


def path(*args, **kwargs):
    p = os.path.join(*args)

    accepted_kwargs = ['create']
    for k in kwargs.keys():
        if k not in accepted_kwargs:
            raise TypeError("path() got an unexpected keyword argument '{}'".format(k))

    if not os.path.exists(p) and kwargs.get('create', False):
        os.makedirs(p)

    return p


def project_path(*args, **kwargs):
    return path(*args, **kwargs)


def which(program):
    import os

    def is_exe(file_path):
        return os.path.isfile(file_path) and os.access(file_path, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for pathname in os.environ["PATH"].split(os.pathsep):
            pathname = pathname.strip('"')
            exe_file = os.path.join(pathname, program)

            if is_exe(exe_file):
                return exe_file

    return None


def remove_and_warn_ifexists(fnm):
    if os.path.exists(fnm):
        logging.warning("Overwritting {}".format(fnm))
        os.remove(fnm)





#!/usr/bin/env python
# -*- coding: utf-8 -*-
import collections

import copy
import sys
import json

import numpy


def _combine_dicts(*args):
    all_dict = []
    for arg in args:
        all_dict += list(arg.items())

    return dict(all_dict)


class Operation:
    def __init__(self, op_code, index, from_symbol, to_symbol):
        self.op_code = op_code
        self.index = index
        self.from_symbol = from_symbol
        self.to_symbol = to_symbol

    def to_primitive(self):
        return self.__dict__

    def __repr__(self):
        return u'({}, {}, {}, {})'.format(self.op_code, self.index, self.from_symbol, self.to_symbol)


class Script:
    def __init__(self, i, j, path=None):
        self.operations = copy.copy(path.operations) if path is not None else []

        self.i = i
        self.j = j

        # The path cannot go any further
        if self.i < 0:
            raise ValueError('Path with a negative index')

    def finished(self):
        return self.i == 0 and self.j == 0

    def to_primitive(self):
        return [op.to_primitive() for op in reversed(self.operations)]

    def to_json(self):
        return json.dumps(self.to_primitive())

    def to_strings(self, use_colors=True):
        """Convert an edit script to a pair of strings representing the operation in a human readable way.

        :param use_colors: Boolean indicating whether to use terminal color codes to color the output.
        :return: Tuple with text corresponding to the first pronunciation and the text of the second one.
        """

        edit_script = self.to_primitive()

        colors = collections.defaultdict(str)

        if use_colors:
            colors['red'] = '\x1b[31m'
            colors['normal'] = '\x1b[m'
            colors['green'] = '\x1b[32m'
            colors['on_red'] = '\x1b[41m'

        src_txt = ''
        dst_txt = ''
        for op in edit_script:
            if op['op_code'] == 'match':
                width = max(len(op['from_symbol']), len(op['to_symbol']))
                if op['from_symbol'] == op['to_symbol']:
                    src_txt += u'{green}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                                 op,
                                                                                                 {'width': width}))
                    dst_txt += u'{green}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                               op,
                                                                                               {'width': width}))
                else:
                    src_txt += u'{red}{from_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                               op,
                                                                                               {'width': width}))
                    dst_txt += u'{red}{to_symbol: ^{width}}{normal}'.format(**_combine_dicts(colors,
                                                                                             op,
                                                                                             {'width': width}))

            elif op['op_code'] == 'insert':
                space = ' '*len(op['to_symbol'])
                src_txt += u'{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors,  op))
                dst_txt += u'{red}{to_symbol}{normal}'.format(**_combine_dicts(colors, op))

            elif op['op_code'] == 'delete':
                space = ' '*len(op['from_symbol'])
                src_txt += u'{red}{from_symbol}{normal}'.format(**_combine_dicts(colors, op))
                dst_txt += u'{on_red}{space}{normal}'.format(space=space, **_combine_dicts(colors, op))

            elif op['op_code'] == 'noninsert':
                continue

            src_txt += ' '
            dst_txt += ' '

        return src_txt, dst_txt

    def print_colors(self):
        src_txt, dst_txt = self.to_strings()
        print(u'---\n{}\n{}\n---'.format(src_txt, dst_txt))


def print_matrix(array, row_labels, col_labels, print_value=lambda x: '{}'.format(x)):
    import pandas
    printit = numpy.vectorize(print_value)
    df = pandas.DataFrame(printit(array), index=row_labels, columns=col_labels)
    print(df)


def print_state(src, trg, ops, costs):
    print_matrix(costs, src, trg)
    print('\n\n')
    print_matrix(ops, src, trg, print_value=lambda x: '({}{}{})'.format('m' if x[0] else ' ',
                                                                        'i' if x[1] else ' ',
                                                                        'd' if x[2] else ' '))


def check_insert(s, t, costs, ops, i, j, cost_method):
    _ = s
    cost = costs[i, j-1] + cost_method(t[j])
    if cost <= costs[i, j]:
        if cost < costs[i, j]:
            ops[i, j] = (False, False, False)

        costs[i, j] = cost
        ops[i, j]['insert'] = True


def check_delete(s, t, costs, ops, i, j, cost_method):
    _ = t
    cost = costs[i-1, j] + cost_method(s[i])
    if cost <= costs[i, j]:
        if cost < costs[i, j]:
            ops[i, j] = (False, False, False)

        costs[i, j] = cost
        ops[i, j]['delete'] = True


def check_match(s, t, costs, ops, i, j, cost_method):
    cost = costs[i-1, j-1] + cost_method(s[i], t[j])
    if cost <= costs[i, j]:
        if cost < costs[i, j]:
            ops[i, j] = (False, False, False)

        costs[i, j] = cost
        ops[i, j]['match'] = True

class TERMINAL(object):
    pass

def best_transforms(src, trg, op_costs=None):
    default_costs = {'match': lambda x, y: 0 if x == y else 1,
                     'insert': lambda x: 1,
                     'delete': lambda x: 1}

    op_costs = op_costs or default_costs

    insert_cost_method = op_costs['insert']
    match_cost_method = op_costs['match']
    delete_cost_method = op_costs['delete']

    src_len = len(src) + 2
    trg_len = len(trg) + 2

    # Initialize costs
    costs = numpy.ones((src_len, trg_len)) * numpy.inf

    op_type = numpy.dtype({'names': ['match',
                                     'insert',
                                     'delete'],
                           'formats': ['b',
                                       'b',
                                       'b']})
    ops = numpy.zeros((src_len, trg_len), dtype=op_type)

    #s = numpy.r_[[''], numpy.array(src), ['']]
    #t = numpy.r_[[''], numpy.array(trg), ['']]
    s = [None] + src + [None]
    t = [None] + trg + [None]

    # Set the cost of the initial cell
    costs[0, 0] = 0

    # Compute costs of inserting up to the nth target item before source
    i = 0
    for j in xrange(1, costs.shape[1]-1):
        check_insert(s, t, costs, ops, i, j, insert_cost_method)

    # Compute costs of deleting up to the nth source item before target
    j = 0
    for i in xrange(1, costs.shape[0]-1):
        check_delete(s, t, costs, ops, i, j, delete_cost_method)

    # Compute costs of everything else
    for i in xrange(1, costs.shape[0]-1):
        for j in xrange(1, costs.shape[1]-1):
            # Replace cost
            check_match(s, t, costs, ops, i, j, match_cost_method)

            # Delete cost
            check_delete(s, t, costs, ops, i, j, delete_cost_method)

            # Insert cost
            check_insert(s, t, costs, ops, i, j, insert_cost_method)

    # Compute costs of inserting up to the nth target item after the source
    i = costs.shape[0]-1
    for j in xrange(1, costs.shape[1]-1):
        check_insert(s, t, costs, ops, i, j, insert_cost_method)

    # Compute costs of deleting up to the nth source item after the target
    j = costs.shape[1]-1
    for i in xrange(1, costs.shape[0]-1):
        check_delete(s, t, costs, ops, i, j, delete_cost_method)

    # Last match cost
    # This last element does not require a cost
    # since it is matching end-of-seq to end-of-seq
    i = costs.shape[0]-1
    j = costs.shape[1]-1
    check_match(s, t, costs, ops, i, j, lambda x, y: 0)

    # Debug the state of the DP algorithm
    # print_state(s, t, ops, costs)

    # Compute transforms by backtracking paths
    paths = []

    if ops[-1, -1]['match']:
        paths.append(Script(src_len - 2, trg_len - 2))

    if ops[-1, -1]['insert']:
        paths.append(Script(src_len - 1, trg_len - 2))

    if ops[-1, -1]['delete']:
        paths.append(Script(src_len - 2, trg_len - 1))

    while any(map(lambda x: not x.finished(), paths)):
        new_paths = []

        for path in paths:
            # If we are in the top left corner of the cost matrix we are finished
            if path.finished():
                new_paths.append(path)
                continue

            if ops[path.i, path.j]['match']:
                op = Operation('match', (path.i-1) * 2 + 1, src[path.i-1], trg[path.j-1])
                new_path = Script(path.i - 1, path.j - 1, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

            if ops[path.i, path.j]['delete']:
                op = Operation('delete', (path.i-1) * 2 + 1, src[path.i-1], None)
                new_path = Script(path.i - 1, path.j, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

            if ops[path.i, path.j]['insert']:
                op = Operation('insert', (path.i-1) * 2 + 2, None, trg[path.j-1])
                new_path = Script(path.i, path.j - 1, path=path)
                new_path.operations.append(op)
                new_paths.append(new_path)

        paths = new_paths

    return costs[i, j], paths, costs, ops


def main():
    a = u'm u ch a'
    b = u'm u ch o s'

    a = u'k w i ð̞ a'
    b = u'i ð̞ a'

    a = u'x e s t o'
    b = u'p ɾ i n θ e s a'

    a = u'p e k ˈe ɲ o s'
    b = u'p e k ˈe ɲ a s'

    a = u'k ˈo ð o'
    b = u'θ ˌe  θ e ð ˌi ʝ a  d ˌe ˌɛ f e θ ˌe  θ e ð ˈi ʝ a'

    if len(sys.argv) > 2:
        a = sys.argv[1]
        b = sys.argv[2]

    src = a.split()
    trg = b.split()

    distance, transfs, costs, ops = best_transforms(src, trg)


    print('distance: {}'.format(distance))
    print('costs:\n{}'.format(costs))
    print('ops:\n{}'.format(ops))
    print(transfs[0].to_json())

    for transf in transfs:
        transf.print_colors()

    print(json.dumps(transfs[0].to_primitive(), indent=4))
    transfs[0].print_colors()

if __name__ == '__main__':
    main()


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# CAVEAT UTILITOR
#
# This file was automatically generated by Grako.
#
#    https://pypi.python.org/pypi/grako/
#
# Any changes you make to it will be overwritten the next time
# the file is generated.


from __future__ import print_function, division, absolute_import, unicode_literals
from grako.parsing import graken, Parser


__version__ = (2014, 10, 21, 13, 45, 6, 1)

__all__ = [
    'htk_modelParser',
    'htk_modelSemantics',
    'main'
]


class htk_modelParser(Parser):
    def __init__(self, whitespace=None, nameguard=True, **kwargs):
        super(htk_modelParser, self).__init__(
            whitespace=whitespace,
            nameguard=nameguard,
            **kwargs
        )

    @graken()
    def _model_(self):

        def block0():
            self._macrodef_()
            self.ast.setlist('macros', self.last_node)
        self._closure(block0)

        def block2():
            self._hmmmacro_()
            self.ast.setlist('hmms', self.last_node)
        self._closure(block2)

        self.ast._define(
            [],
            ['macros', 'hmms']
        )

    @graken()
    def _macrodef_(self):
        with self._choice():
            with self._option():
                self._transPmacro_()
                self.ast['transition'] = self.last_node
            with self._option():
                self._stateinfomacro_()
                self.ast['state'] = self.last_node
            with self._option():
                self._optmacro_()
                self.ast['options'] = self.last_node
            with self._option():
                self._varmacro_()
                self.ast['variance'] = self.last_node
            with self._option():
                self._meanmacro_()
                self.ast['mean'] = self.last_node
            with self._option():
                self._durationmacro_()
                self.ast['duration'] = self.last_node
            self._error('no available options')

        self.ast._define(
            ['transition', 'state', 'options', 'variance', 'mean', 'duration'],
            []
        )

    @graken()
    def _hmmmacro_(self):
        with self._optional():
            self._hmmref_()
            self.ast['name'] = self.last_node
        self._hmmdef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _optmacro_(self):
        self._token('~o')
        self._cut()
        self._globalOpts_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['definition'],
            []
        )

    @graken()
    def _transPmacro_(self):
        self._transPref_()
        self.ast['name'] = self.last_node
        self._transPdef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _stateinfomacro_(self):
        self._stateinforef_()
        self.ast['name'] = self.last_node
        self._stateinfodef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _varmacro_(self):
        self._varref_()
        self.ast['name'] = self.last_node
        self._vardef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _meanmacro_(self):
        self._meanref_()
        self.ast['name'] = self.last_node
        self._meandef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _durationmacro_(self):
        self._durationref_()
        self.ast['name'] = self.last_node
        self._durationdef_()
        self.ast['definition'] = self.last_node

        self.ast._define(
            ['name', 'definition'],
            []
        )

    @graken()
    def _varref_(self):
        self._token('~v')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _transPref_(self):
        self._token('~t')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _stateinforef_(self):
        self._token('~s')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _hmmref_(self):
        self._token('~h')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _weightsref_(self):
        self._token('~w')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _mixpdfref_(self):
        self._token('~m')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _meanref_(self):
        self._token('~u')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _durationref_(self):
        self._token('~d')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _invref_(self):
        self._token('~i')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _xformref_(self):
        self._token('~x')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _inputXformref_(self):
        self._token('~j')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node

    @graken()
    def _macro_(self):
        self._string_()

    @graken()
    def _hmmdef_(self):
        self._token('<BeginHMM>')
        self._cut()
        with self._optional():
            self._globalOpts_()
            self.ast['options'] = self.last_node
        self._token('<NumStates>')
        self._cut()
        self._short_()
        self.ast['state_count'] = self.last_node

        def block2():
            self._state_()
            self.ast.setlist('states', self.last_node)
        self._positive_closure(block2)

        with self._optional():
            self._regTree_()
            self.ast['regression_tree'] = self.last_node
        self._transP_()
        self.ast['transition'] = self.last_node
        with self._optional():
            self._duration_()
            self.ast['duration'] = self.last_node
        self._token('<EndHMM>')

        self.ast._define(
            ['options', 'state_count', 'regression_tree', 'transition', 'duration'],
            ['states']
        )

    @graken()
    def _globalOpts_(self):

        def block1():
            self._option_()
        self._positive_closure(block1)

        self.ast['@'] = self.last_node

    @graken()
    def _option_(self):
        with self._choice():
            with self._option():
                self._token('<HmmSetId>')
                self._cut()
                self._string_()
                self.ast['hmm_set_id'] = self.last_node
            with self._option():
                self._token('<StreamInfo>')
                self._cut()
                self._streaminfo_()
                self.ast['stream_info'] = self.last_node
            with self._option():
                self._token('<VecSize>')
                self._cut()
                self._short_()
                self.ast['vector_size'] = self.last_node
            with self._option():
                self._token('<InputXform>')
                self._cut()
                self._inputXform_()
                self.ast['input_transform'] = self.last_node
            with self._option():
                self._covkind_()
                self.ast['covariance_kind'] = self.last_node
            with self._option():
                self._durkind_()
                self.ast['duration_kind'] = self.last_node
            with self._option():
                self._parmkind_()
                self.ast['parameter_kind'] = self.last_node
            self._error('no available options')

        self.ast._define(
            ['hmm_set_id', 'stream_info', 'vector_size', 'input_transform', 'covariance_kind', 'duration_kind', 'parameter_kind'],
            []
        )

    @graken()
    def _streaminfo_(self):
        self._short_()
        self.ast['count'] = self.last_node

        def block2():
            self._short_()
        self._closure(block2)
        self.ast['sizes'] = self.last_node

        self.ast._define(
            ['count', 'sizes'],
            []
        )

    @graken()
    def _covkind_(self):
        self._token('<')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('diagc')
                with self._option():
                    self._token('invdiagc')
                with self._option():
                    self._token('fullc')
                with self._option():
                    self._token('lltc')
                with self._option():
                    self._token('xformc')
                self._error('expecting one of: diagc fullc invdiagc lltc xformc')
        self.ast['@'] = self.last_node
        self._token('>')

    @graken()
    def _durkind_(self):
        self._token('<')
        with self._group():
            with self._choice():
                with self._option():
                    self._token('nulld')
                with self._option():
                    self._token('poissond')
                with self._option():
                    self._token('gammad')
                with self._option():
                    self._token('gen')
                self._error('expecting one of: gammad gen nulld poissond')
        self.ast['@'] = self.last_node
        self._token('>')

    @graken()
    def _parmkind_(self):
        self._token('<')
        self._basekind_()
        self.ast['base'] = self.last_node

        def block2():
            with self._choice():
                with self._option():
                    self._token('_D')
                with self._option():
                    self._token('_A')
                with self._option():
                    self._token('_T')
                with self._option():
                    self._token('_E')
                with self._option():
                    self._token('_N')
                with self._option():
                    self._token('_Z')
                with self._option():
                    self._token('_O')
                with self._option():
                    self._token('_V')
                with self._option():
                    self._token('_C')
                with self._option():
                    self._token('_K')
                self._error('expecting one of: _A _C _D _E _K _N _O _T _V _Z')
        self._closure(block2)
        self.ast['options'] = self.last_node
        self._token('>')

        self.ast._define(
            ['base', 'options'],
            []
        )

    @graken()
    def _basekind_(self):
        with self._choice():
            with self._option():
                self._token('discrete')
            with self._option():
                self._token('lpc')
            with self._option():
                self._token('lpcepstra')
            with self._option():
                self._token('mfcc')
            with self._option():
                self._token('fbank')
            with self._option():
                self._token('melspec')
            with self._option():
                self._token('lprefc')
            with self._option():
                self._token('lpdelcep')
            with self._option():
                self._token('user')
            self._error('expecting one of: discrete fbank lpc lpcepstra lpdelcep lprefc melspec mfcc user')

    @graken()
    def _state_(self):
        self._token('<State>')
        self._cut()
        self._short_()
        self.ast['index'] = self.last_node
        self._stateinfo_()
        self.ast['state'] = self.last_node

        self.ast._define(
            ['index', 'state'],
            []
        )

    @graken()
    def _stateinfo_(self):
        with self._choice():
            with self._option():
                self._stateinforef_()
                self.ast['@'] = self.last_node
            with self._option():
                self._stateinfodef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _stateinfodef_(self):
        with self._optional():
            self._mixes_()
            self.ast['streams_mixcount'] = self.last_node
        with self._optional():
            self._weights_()
            self.ast['weights'] = self.last_node

        def block2():
            self._stream_()
            self.ast.setlist('streams', self.last_node)
        self._positive_closure(block2)

        with self._optional():
            self._duration_()
            self.ast['duration'] = self.last_node

        self.ast._define(
            ['streams_mixcount', 'weights', 'duration'],
            ['streams']
        )

    @graken()
    def _mixes_(self):
        self._token('<NumMixes>')
        self._cut()

        def block1():
            self._short_()
        self._positive_closure(block1)

        self.ast['@'] = self.last_node

    @graken()
    def _weights_(self):
        with self._choice():
            with self._option():
                self._weightsref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._weightsdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _weightsdef_(self):
        self._token('<SWeights>')
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._vector_()
        self.ast['vector'] = self.last_node

        self.ast._define(
            ['dim', 'vector'],
            []
        )

    @graken()
    def _stream_(self):
        with self._optional():
            self._token('<Stream>')
            self._cut()
            self._short_()
            self.ast['dim'] = self.last_node
        with self._group():
            with self._choice():
                with self._option():

                    def block1():
                        self._mixture_()
                        self.ast.setlist('mixtures', self.last_node)
                    self._positive_closure(block1)
                with self._option():
                    self._tmixpdf_()
                    self.ast['tmixpdf'] = self.last_node
                with self._option():
                    self._discpdf_()
                    self.ast['discpdf'] = self.last_node
                self._error('no available options')

        self.ast._define(
            ['dim', 'tmixpdf', 'discpdf'],
            ['mixtures']
        )

    @graken()
    def _mixture_(self):
        with self._optional():
            self._token('<Mixture>')
            self._cut()
            self._short_()
            self.ast['index'] = self.last_node
            self._float_()
            self.ast['weight'] = self.last_node
        self._mixpdf_()
        self.ast['pdf'] = self.last_node

        self.ast._define(
            ['index', 'weight', 'pdf'],
            []
        )

    @graken()
    def _tmixpdf_(self):
        self._token('<TMix>')
        self._cut()
        self._macro_()
        self._weightList_()

    @graken()
    def _weightList_(self):

        def block0():
            self._repShort_()
        self._positive_closure(block0)

    @graken()
    def _repShort_(self):
        self._short_()
        with self._optional():
            self._token('*')
            self._cut()
            self._char_()

    @graken()
    def _discpdf_(self):
        self._token('<DProb>')
        self._cut()
        self._weightList_()

    @graken()
    def _mixpdf_(self):
        with self._choice():
            with self._option():
                self._mixpdfref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._mixpdfdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _mixpdfdef_(self):
        with self._optional():
            self._rclass_()
            self.ast['regression_class'] = self.last_node
        self._mean_()
        self.ast['mean'] = self.last_node
        self._cov_()
        self.ast['covariance'] = self.last_node
        with self._optional():
            self._token('<GConst>')
            self._cut()
            self._float_()
            self.ast['gconst'] = self.last_node

        self.ast._define(
            ['regression_class', 'mean', 'covariance', 'gconst'],
            []
        )

    @graken()
    def _rclass_(self):
        self._token('<RClass>')
        self._cut()
        self._short_()
        self.ast['@'] = self.last_node

    @graken()
    def _mean_(self):
        with self._choice():
            with self._option():
                self._meanref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._meandef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _meandef_(self):
        self._token('<Mean>')
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._vector_()
        self.ast['vector'] = self.last_node

        self.ast._define(
            ['dim', 'vector'],
            []
        )

    @graken()
    def _cov_(self):
        with self._choice():
            with self._option():
                self._var_()
                self.ast['variance'] = self.last_node
            with self._option():
                self._inv_()
            with self._option():
                self._xform_()
            self._error('no available options')

        self.ast._define(
            ['variance'],
            []
        )

    @graken()
    def _var_(self):
        with self._choice():
            with self._option():
                self._varref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._vardef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _vardef_(self):
        self._token('<Variance>')
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._vector_()
        self.ast['vector'] = self.last_node

        self.ast._define(
            ['dim', 'vector'],
            []
        )

    @graken()
    def _inv_(self):
        with self._choice():
            with self._option():
                self._invref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._invdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _invdef_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('<InvCovar>')
                with self._option():
                    self._token('<LLTCovar>')
                self._error('expecting one of: <InvCovar> <LLTCovar>')
        self.ast['type'] = self.last_node
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._tmatrix_()
        self.ast['matrix'] = self.last_node

        self.ast._define(
            ['type', 'dim', 'matrix'],
            []
        )

    @graken()
    def _xform_(self):
        with self._choice():
            with self._option():
                self._xformref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._xformdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _xformdef_(self):
        self._token('<Xform>')
        self._cut()
        self._short_()
        self.ast['dim1'] = self.last_node
        self._short_()
        self.ast['dim2'] = self.last_node
        self._matrix_()
        self.ast['matrix'] = self.last_node

        self.ast._define(
            ['dim1', 'dim2', 'matrix'],
            []
        )

    @graken()
    def _tmatrix_(self):
        self._matrix_()

    @graken()
    def _duration_(self):
        with self._choice():
            with self._option():
                self._durationref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._durationdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _durationdef_(self):
        self._token('<Duration>')
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._vector_()
        self.ast['vector'] = self.last_node

        self.ast._define(
            ['dim', 'vector'],
            []
        )

    @graken()
    def _regTree_(self):
        self._token('~r')
        self._cut()
        self._macro_()
        self.ast['@'] = self.last_node
        self._tree_()
        self.ast['tree'] = self.last_node

        self.ast._define(
            ['tree'],
            []
        )

    @graken()
    def _tree_(self):
        self._token('<RegTree>')
        self._cut()
        self._short_()
        self._nodes_()

    @graken()
    def _nodes_(self):
        with self._group():
            with self._choice():
                with self._option():
                    self._token('<Node>')
                    self._cut()
                    self._short_()
                    self._short_()
                    self._short_()
                with self._option():
                    self._token('<TNode>')
                    self._cut()
                    self._short_()
                    self._int_()
                self._error('no available options')
        with self._optional():
            self._nodes_()

    @graken()
    def _transP_(self):
        with self._choice():
            with self._option():
                self._transPref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._transPdef_()
                self.ast['@'] = self.last_node
            self._error('no available options')

    @graken()
    def _transPdef_(self):
        self._token('<TransP>')
        self._cut()
        self._short_()
        self.ast['dim'] = self.last_node
        self._matrix_()
        self.ast['array'] = self.last_node

        self.ast._define(
            ['dim', 'array'],
            []
        )

    @graken()
    def _inputXform_(self):
        with self._choice():
            with self._option():
                self._inputXformref_()
                self.ast['@'] = self.last_node
            with self._option():
                self._inhead_()
                self._inmatrix_()
            self._error('no available options')

    @graken()
    def _inhead_(self):
        self._token('<MMFIdMask>')
        self._cut()
        self._string_()
        self._parmkind_()
        with self._optional():
            self._token('<PreQual>')

    @graken()
    def _inmatrix_(self):
        self._token('<LinXform>')
        self._token('<VecSize>')
        self._cut()
        self._short_()
        self._token('<BlockInfo>')
        self._cut()
        self._short_()

        def block0():
            self._short_()
        self._positive_closure(block0)


        def block1():
            self._block_()
        self._positive_closure(block1)

    @graken()
    def _block_(self):
        self._token('<Block>')
        self._cut()
        self._short_()
        self._xform_()

    @graken()
    def _string_(self):
        self._pattern(r'.*')

    @graken()
    def _vector_(self):
        self._pattern(r'[\d.\-\+eE \n]+')

    @graken()
    def _matrix_(self):
        self._pattern(r'[\d.\-\+eE \n]+')

    @graken()
    def _short_(self):
        self._pattern(r'\d+')

    @graken()
    def _float_(self):
        self._pattern(r'[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?')

    @graken()
    def _char_(self):
        self._pattern(r'.')

    @graken()
    def _int_(self):
        self._pattern(r'[-+]?(0[xX][\dA-Fa-f]+|0[0-7]*|\d+)')


class htk_modelSemantics(object):
    def model(self, ast):
        return ast

    def macrodef(self, ast):
        return ast

    def hmmmacro(self, ast):
        return ast

    def optmacro(self, ast):
        return ast

    def transPmacro(self, ast):
        return ast

    def stateinfomacro(self, ast):
        return ast

    def varmacro(self, ast):
        return ast

    def meanmacro(self, ast):
        return ast

    def durationmacro(self, ast):
        return ast

    def varref(self, ast):
        return ast

    def transPref(self, ast):
        return ast

    def stateinforef(self, ast):
        return ast

    def hmmref(self, ast):
        return ast

    def weightsref(self, ast):
        return ast

    def mixpdfref(self, ast):
        return ast

    def meanref(self, ast):
        return ast

    def durationref(self, ast):
        return ast

    def invref(self, ast):
        return ast

    def xformref(self, ast):
        return ast

    def inputXformref(self, ast):
        return ast

    def macro(self, ast):
        return ast

    def hmmdef(self, ast):
        return ast

    def globalOpts(self, ast):
        return ast

    def option(self, ast):
        return ast

    def streaminfo(self, ast):
        return ast

    def covkind(self, ast):
        return ast

    def durkind(self, ast):
        return ast

    def parmkind(self, ast):
        return ast

    def basekind(self, ast):
        return ast

    def state(self, ast):
        return ast

    def stateinfo(self, ast):
        return ast

    def stateinfodef(self, ast):
        return ast

    def mixes(self, ast):
        return ast

    def weights(self, ast):
        return ast

    def weightsdef(self, ast):
        return ast

    def stream(self, ast):
        return ast

    def mixture(self, ast):
        return ast

    def tmixpdf(self, ast):
        return ast

    def weightList(self, ast):
        return ast

    def repShort(self, ast):
        return ast

    def discpdf(self, ast):
        return ast

    def mixpdf(self, ast):
        return ast

    def mixpdfdef(self, ast):
        return ast

    def rclass(self, ast):
        return ast

    def mean(self, ast):
        return ast

    def meandef(self, ast):
        return ast

    def cov(self, ast):
        return ast

    def var(self, ast):
        return ast

    def vardef(self, ast):
        return ast

    def inv(self, ast):
        return ast

    def invdef(self, ast):
        return ast

    def xform(self, ast):
        return ast

    def xformdef(self, ast):
        return ast

    def tmatrix(self, ast):
        return ast

    def duration(self, ast):
        return ast

    def durationdef(self, ast):
        return ast

    def regTree(self, ast):
        return ast

    def tree(self, ast):
        return ast

    def nodes(self, ast):
        return ast

    def transP(self, ast):
        return ast

    def transPdef(self, ast):
        return ast

    def inputXform(self, ast):
        return ast

    def inhead(self, ast):
        return ast

    def inmatrix(self, ast):
        return ast

    def block(self, ast):
        return ast

    def string(self, ast):
        return ast

    def vector(self, ast):
        return ast

    def matrix(self, ast):
        return ast

    def short(self, ast):
        return ast

    def float(self, ast):
        return ast

    def char(self, ast):
        return ast

    def int(self, ast):
        return ast


def main(filename, startrule, trace=False, whitespace=None):
    import json
    with open(filename) as f:
        text = f.read()
    parser = htk_modelParser(parseinfo=False)
    ast = parser.parse(
        text,
        startrule,
        filename=filename,
        trace=trace,
        whitespace=whitespace)
    print('AST:')
    print(ast)
    print()
    print('JSON:')
    print(json.dumps(ast, indent=2))
    print()

if __name__ == '__main__':
    import argparse
    import string
    import sys

    class ListRules(argparse.Action):
        def __call__(self, parser, namespace, values, option_string):
            print('Rules:')
            for r in htk_modelParser.rule_list():
                print(r)
            print()
            sys.exit(0)

    parser = argparse.ArgumentParser(description="Simple parser for htk_model.")
    parser.add_argument('-l', '--list', action=ListRules, nargs=0,
                        help="list all rules and exit")
    parser.add_argument('-t', '--trace', action='store_true',
                        help="output trace information")
    parser.add_argument('-w', '--whitespace', type=str, default=string.whitespace,
                        help="whitespace specification")
    parser.add_argument('file', metavar="FILE", help="the input file to parse")
    parser.add_argument('startrule', metavar="STARTRULE",
                        help="the start rule for parsing")
    args = parser.parse_args()

    main(
        args.file,
        args.startrule,
        trace=args.trace,
        whitespace=args.whitespace
    )



#!/usr/bin/env python
"""load_htk_model
Reads an HTK model file

Usage:
    load_htk_model <model_file>... [options]
    load_htk_model --help

Options:
    <model_file>                                    the files of the model (e.g. macros hmmdefs)
    -o <output_file>, --output_file <output_file>   the output file (default is stdout)
    -h, --help                                      print this help screen
"""

from __future__ import absolute_import, division, print_function, unicode_literals
import copy

__author__ = 'rmarxer'


import numpy as np
import collections

from . import htk_model


def create_default():
    return collections.defaultdict(lambda: None)


def create_vector(vector):
    return {'dim': vector.size, 'vector': vector}


def create_square_matrix(mat):
    return {'dim': mat.shape[0], 'matrix': mat}


def create_transition(state_stay_probabilites=[0.6, 0.6, 0.7]):
    state_probs = np.array(state_stay_probabilites)
    n_states = state_probs.size + 2
    transitions = np.zeros((n_states, n_states))
    transitions[0, 1] = 1.0
    for i, p in enumerate(state_probs):
        transitions[i + 1, i + 1] = p
        transitions[i + 1, i + 2] = 1 - p

    return create_square_matrix(transitions)


def create_parameter_kind(base=None, options=[]):
    result = create_default()
    result['base'] = base
    result['options'] = options
    return result


def create_options(vector_size=None,
                   parameter_kind=None):

    macro = create_default()

    options = []
    if vector_size:
        option = create_default()
        option['vector_size'] = vector_size
        options.append(option)

    if parameter_kind:
        option = create_default()
        option['parameter_kind'] = parameter_kind
        options.append(option)

    macro['options'] = {'definition': options}

    return macro


def create_gmm(means, variances, gconsts=None, weights=None):
    mixtures = []

    if means.ndim == 1:
        means = means[None, :]
        variances = variances[None, :]

    gmm = create_default()

    for i in range(means.shape[0]):
        mixture = create_default()
        mixture['pdf'] = create_default()
        mixture['pdf']['mean'] = create_vector(means[i])
        mixture['pdf']['covariance'] = create_default()
        mixture['pdf']['covariance']['variance'] = create_vector(variances[i])

        if gconsts is not None:
            mixture['pdf']['gconst'] = gconsts[i]

        if weights is not None:
            mixture['weight'] = weights[i]

        mixtures.append(mixture)

    stream = create_default()
    stream['mixtures'] = mixtures

    gmm['streams'] = [stream]

    return gmm


def create_hmm(states, transition, name=None):
    hmm = create_default()
    hmm['name'] = name
    hmm['definition'] = create_default()

    hmm['definition']['state_count'] = len(states) + 2
    hmm['definition']['states'] = []
    for i, state in enumerate(states):
        hmm_state = create_default()

        hmm_state['index'] = i + 2
        hmm_state['state'] = state

        hmm['definition']['states'].append(hmm_state)

    hmm['definition']['transition'] = transition

    return hmm


def create_model(macros, hmms):
    model = create_default()

    model['macros'] = macros
    model['hmms'] = hmms

    return model


def create_prototype(sample_dimension,
                     parameter_kind_base='user',
                     parameter_kind_options=[],
                     state_stay_probabilities=[0.6, 0.6, 0.7]):
    """Create a prototype HTK model file using a feature file.
    """
    parameter_kind = create_parameter_kind(base=parameter_kind_base,
                                           options=parameter_kind_options)

    transition = create_transition(state_stay_probabilities)

    state_count = len(state_stay_probabilities)

    states = []
    for i in range(state_count):
        state = create_gmm(np.zeros(sample_dimension),
                           np.ones(sample_dimension),
                           weights=None,
                           gconsts=None)

        states.append(state)

    hmms = [create_hmm(states, transition)]
    macros = [create_options(vector_size=sample_dimension,
                             parameter_kind=parameter_kind)]

    model = create_model(macros, hmms)

    return model


def split_model(model):
    macros_model = copy.deepcopy(model)
    hmmdefs_model = copy.deepcopy(model)

    filter_macros = lambda x: x['transition'] is None and x['state'] is None

    macros_model['macros'] = filter(filter_macros, model['macros'])
    macros_model['hmms'] = []

    hmmdefs_model['macros'] = filter(lambda x: not filter_macros(x), model['macros'])

    return macros_model, hmmdefs_model


def map_hmms(input_model, mapping):
    """Create a new HTK HMM model given a model and a mapping dictionary.

    :param input_model: The model to transform of type dict
    :param mapping: A dictionary from string -> list(string)
    :return: The transformed model of type dict
    """

    output_model = copy.copy(input_model)

    o_hmms = []
    for i_hmm in input_model['hmms']:
        i_hmm_name = i_hmm['name']
        o_hmm_names = mapping.get(i_hmm_name, [i_hmm_name])

        for o_hmm_name in o_hmm_names:
            o_hmm = copy.copy(i_hmm)
            o_hmm['name'] = o_hmm_name

            o_hmms.append(o_hmm)

    output_model['hmms'] = o_hmms

    return output_model


def transform_to_semi_hmm(model, repetition_count=4, min_repetitions=1, max_repetitions=np.inf):
    """This function expands and HMM model into a semi HMM model by multiplying the number of states repetition_coutn:
    times and modify the transition matrices accordingly to only allow between min_repetitions: and max_repetitions:
    repetitions.

    :param model: The model to transform in place
    :param repetition_count: The amount of repeated occupations of a state to model
    :param min_repetitions: The minimum amount of repeated occupations of a state to allow
    :param max_repetitions: The maximum amount of repeated occupations of a state to allow
    """

    def transform_transition(definition):
        """Transform in place the definition of the transition element.
        :param definition: The transition element
        """
        dim = definition['dim']
        matrix = definition['matrix']

        new_dim = (definition['dim'] - 2) * repetition_count + 2

        new_matrix = np.zeros((new_dim, new_dim))

        new_matrix[0, 1] = matrix[0, 1]

        def state_repeat_to_ind(sta, rep):
            """Return the index corresponding to the state sta: and repetition rep:.
            States and repetitions start at 0 (state 0 corresponds to the initial dummy state).

            :param sta: The state for which to get the index.
            :param rep: The repetition for which to get the index
            :return: The index in the transition matrix
            """
            if sta == 0:
                return 0

            return ((sta - 1) * repetition_count + rep) + 1

        def probability(source_state, source_repeat, target_state, target_repeat):
            """Return the computed probability of the transition,
            given the source (state source_state:, repeat source_repeat:)
            and the target (state target_state:, repeat target_repeat:).
            Currently this simply keeps the probabilistic model unchanged.

            :param source_state: The source state
            :param source_repeat: The source repetition
            :param target_state: The target state
            :param target_repeat: The target repetition
            """
            return matrix[source_state, target_state]

        for state in range(1, dim-1):
            for repeat in range(min_repetitions-1, min(max_repetitions, repetition_count)):
                source = state_repeat_to_ind(state, repeat)

                # Case in which we advance to the next state
                target = state_repeat_to_ind(state+1, 0)
                prob = probability(state, repeat, state+1, 0)
                new_matrix[source, target] = prob

                if repeat == repetition_count - 1:
                    if np.isinf(max_repetitions):
                        # Case in which we repeat the last state that we model (self-transition)
                        target = source
                        prob = probability(state, repeat, state, repeat)
                        new_matrix[source, target] = prob

                else:
                    # Case in which repeat the state again
                    target = state_repeat_to_ind(state, repeat+1)
                    prob = probability(state, repeat, state, repeat+1)
                    new_matrix[source, target] = prob

        new_matrix[:-1, :] /= new_matrix.sum(axis=1)[:-1, np.newaxis]

        definition['dim'] = new_dim
        definition['matrix'] = new_matrix
        return definition

    # Go through all hmm models
    transition_references = set()
    for hmm in model['hmms']:
        new_statecount = (hmm['definition']['state_count'] - 2) * repetition_count + 2
        hmm['definition']['state_count'] = new_statecount

        # Expand states mantaining the expanded states tied
        new_states = []
        index = None
        for state in sorted(hmm['definition']['states'], key=lambda x: x['index']):
            index = index or state['index']
            for repetition in range(repetition_count):
                new_state = copy.copy(state)
                new_state['index'] = index
                new_states.append(new_state)
                index += 1

        hmm['definition']['states'] = new_states

        # Transform the transition matrices
        if isinstance(hmm['definition']['transition'], collections.Mapping):
            # Definition: transform now
            transform_transition(hmm['definition']['transition']['definition'])
        else:
            # Reference: collect it and transform them all later
            transition_references.add(hmm['definition']['transition'])

    for reference in transition_references:
        definitions = filter(lambda x: x['transition'] and x['transition']['name'] == reference, model['macros'])

        if not len(definitions) == 1:
            logging.warning('Invalid number of definitions ({}) for transition macro {}.'
                            '  Expecting 1 definition.'.format(len(definitions), reference))

        for definition in definitions:
            transform_transition(definition['transition']['definition'])


def main():
    model = htk_model.load_model('test')
    macros, hmmdefs = split_model(model)
    print(htk_model.model_to_json(macros, indent=4))
    print(htk_model.model_to_json(hmmdefs, indent=4))

if __name__ == '__main__':
    main()


#!/usr/bin/env python
"""htk
Reading and writing HTK files.

(Currently only handles diagonal covariance mixture models)
"""

import numpy as np
import re
from collections import namedtuple
import struct
import codecs
import unicodedata

from . import htk_model

Label = namedtuple('Label', 'start end symbol log_likelihood word')

ParameterMeta = namedtuple('ParameterMeta', 'n_samples samp_period samp_size parm_kind '
                                            'parm_kind_base parm_kind_opts parm_kind_str')
Parameter = namedtuple('Parameter', 'meta samples')

HTK_HYPOTHESIS_PATTERN = r'''
    (\s*
        ((?P<start>\d+)
         [ \t\f\v]?)?
        ((?P<end>\d+)?
         [ \t\f\v]?)?
        ((?P<symbol>\S+)
         [ \t\f\v]?)
        ((?P<log_likelihood>[-+]?\d+\.?\d*)?
         [ \t\f\v]?)?
        ((?P<word>\S+))?
    \s*)
    '''

HTK_MLF_PATTERN = r'''
      (?:^\"(?P<file>.*?)\"$)   # The file name
      (?P<hypotheses>.*?)
      (?:^\.$)
    '''

HTK_HYPOTHESIS_RE = re.compile(HTK_HYPOTHESIS_PATTERN, re.MULTILINE | re.VERBOSE | re.UNICODE)
HTK_MLF_RE = re.compile(HTK_MLF_PATTERN, re.MULTILINE | re.VERBOSE | re.UNICODE | re.DOTALL)

HTK_PARAMETER_KINDS = {
    0: 'WAVEFORM',  # sampled waveform
    1: 'LPC',  # linear prediction filter coefficients
    2: 'LPREFC',  # linear prediction reflection coefficients
    3: 'LPCEPSTRA',  # LPC cepstral coefficients
    4: 'LPDELCEP',  # LPC cepstra plus delta coefficients
    5: 'IREFC',  # LPC reflection coef in 16 bit integer format
    6: 'MFCC',  # mel-frequency cepstral coefficients
    7: 'FBANK',  # log mel-filter bank channel outputs
    8: 'MELSPEC',  # linear mel-filter bank channel outputs
    9: 'USER',  # user defined sample kind
    10: 'DISCRETE',  # vector quantised data
}

HTK_PARAMETER_FLAGS = [
    ('E', 0o000100),  # has energy
    ('N', 0o000200),  # absolute energy suppressed
    ('D', 0o000400),  # has delta coefficients
    ('A', 0o001000),  # has acceleration coefficients
    ('C', 0o002000),  # is compressed
    ('Z', 0o004000),  # has zero mean static coef.
    ('K', 0o010000),  # has CRC checksum
    ('O', 0o020000),  # has 0'th cepstral coef.
    ('V', 0o040000),  # has VQ data
    ('T', 0o100000),  # has third differential coef.
]


def _htk_str_to_param(parm_str):
    tokens = parm_str.split('_')

    inv_htk_parameter_kinds = dict([(v, k) for k, v in HTK_PARAMETER_KINDS.items()])
    dict_htk_parameter_flags = dict(HTK_PARAMETER_FLAGS)

    parm_kind = inv_htk_parameter_kinds[tokens[0]]

    for flag in tokens[1:]:
        parm_kind = dict_htk_parameter_flags[flag]

    return parm_kind


def _htk_param_to_str(parm_kind):
    # first 6 bits are the main parameter kind
    result = HTK_PARAMETER_KINDS[parm_kind & 0b000000111111]

    # check for the flags
    for flag, mask in HTK_PARAMETER_FLAGS:
        if parm_kind & mask:
            result += '_' + flag

    return result


def _htk_param_to_base_options(parm_kind):
    # first 6 bits are the main parameter kind
    base = HTK_PARAMETER_KINDS[parm_kind & 0b000000111111]

    # check for the flags
    options = []
    for flag, mask in HTK_PARAMETER_FLAGS:
        if parm_kind & mask:
            options.append('_' + flag)

    return base, options


HTK_HEADER_FORMAT = '>iihh'


def _parse_parameter_meta(input_file):
    buffer = input_file.read(struct.calcsize(HTK_HEADER_FORMAT))
    n_samples, samp_period, samp_size, parm_kind = struct.unpack_from(HTK_HEADER_FORMAT, buffer)
    parm_kind_str = _htk_param_to_str(parm_kind)
    parm_kind_base, parm_kind_opts = _htk_param_to_base_options(parm_kind)

    if '_C' in parm_kind_str:
        raise NotImplementedError('HTK files with compression flag (_C) not supported.')

    if '_K' in parm_kind_str:
        raise NotImplementedError('HTK files with CRC flag (_K) not supported.')

    meta = ParameterMeta(n_samples=n_samples,
                         samp_period=samp_period,
                         samp_size=samp_size,
                         parm_kind=parm_kind,
                         parm_kind_base=parm_kind_base,
                         parm_kind_opts=parm_kind_opts,
                         parm_kind_str=parm_kind_str)
    return meta


def _parse_parameter_samples(input_file, meta):
    samples = []
    sample_format = '>%df' % (meta.samp_size / 4)
    for i in range(meta.n_samples):
        buffer = input_file.read(struct.calcsize(sample_format))
        sample = struct.unpack_from(sample_format, buffer)
        samples.append(list(sample))

    return Parameter(meta=meta,
                     samples=np.array(samples))


def _parse_parameter(input_file):
    meta = _parse_parameter_meta(input_file)
    return _parse_parameter_samples(input_file, meta)


def _serialize_parameter(parameter, output_file):
    output_file.write(struct.pack(HTK_HEADER_FORMAT,
                                  parameter.meta.n_samples,
                                  parameter.meta.samp_period,
                                  parameter.meta.samp_size,
                                  parameter.meta.parm_kind))

    sample_dimension = parameter.meta.samp_size / 4
    sample_format = '>%df' % (sample_dimension,)
    for i, sample in enumerate(parameter.samples):
        if len(sample) != sample_dimension:
            raise ValueError('All the samples written to an HTK file must have the same size.'
                             'Sample {} has length {}, while previous samples had length {}'.format(i,
                                                                                                    len(sample),
                                                                                                    sample_dimension))

        output_file.write(struct.pack(sample_format, *tuple(sample)))
    return


def load_parameter(input_filename, only_header=False):
    """Load HTK parameter/feature file.

    :param input_filename:
    :param only_header: only load the metadata
    :return: a named tuple representing the HTK parmeter file
    """

    with open(input_filename, 'rb') as f:
        meta = _parse_parameter_meta(f)
        if only_header:
            return meta

        return _parse_parameter_samples(f, meta)


def save_parameter(parameter, output_filename):
    """Save a file in HTK Parameter File Format.

    :param parameter: The named tuple Parameter containing the parameter file metadata and data
    :param output_filename: The name of the file where to save
    """
    with open(output_filename, 'wb') as f:
        _serialize_parameter(parameter, f)


def create_parameter(samples, sample_period):
    """Create a HTK Parameter object from an array of samples and a samples period

    :param samples (list of lists or array of floats): The samples to write into the file. Usually feature vectors.
    :param sample_period (int): Sample period in 100ns units.
    """

    parm_kind_str = 'USER'
    parm_kind = _htk_str_to_param(parm_kind_str)
    parm_kind_base, parm_kind_opts = _htk_str_to_param(parm_kind_str)

    meta = ParameterMeta(n_samples=len(samples),
                         samp_period=sample_period,
                         samp_size=len(samples[0]) * 4,  # size in bytes
                         parm_kind_str=parm_kind_str,
                         parm_kind=parm_kind,
                         parm_kind_base=parm_kind_base,
                         parm_kind_opts=parm_kind_opts)
    return Parameter(meta=meta,
                     samples=np.array(samples))


def load_mlf(filename, utf8_normalization=None):
    """Load an HTK Master Label File.

    :param filename: The filename of the MLF file.
    :param utf8_normalization: None
    """
    with codecs.open(filename, 'r', 'string_escape') as f:
        data = f.read().decode('utf8')
        if utf8_normalization:
            data = unicodedata.normalize(utf8_normalization, data)

    mlfs = {}
    for mlf_object in HTK_MLF_RE.finditer(data):
        mlfs[mlf_object.group('file')] = [[Label(**mo.groupdict())
                                           for mo
                                           in HTK_HYPOTHESIS_RE.finditer(recognition_data)]
                                          for recognition_data
                                          in re.split(r'\n///\n', mlf_object.group('hypotheses'))]

    return mlfs


def save_mlf(mlf, output_filename):
    """Save an HTK Master Label File.

    :param mlf: MLF dictionary containing a mapping from file to list of annotations.
    :param output_filename: The file where to save the MLF
    """
    with codecs.open(output_filename, 'w', 'utf-8') as f:
        f.write(u'#!MLF!#\n')
        for k, v in mlf.items():
            f.write(u'"{}"\n'.format(k))
            for labels in v:
                for label in labels:
                    line = u'{start} {end} {symbol} ' \
                           u'{loglikelihood} {word}'.format(start=label.start or '',
                                                            end=label.end or '',
                                                            symbol=label.symbol or '',
                                                            loglikelihood=label.log_likelihood or '',
                                                            word=label.word or '')
                    f.write(u'{}\n'.format(line.strip()))

                f.write(u'.\n')


def load_model(filename):
    """Load an HTK HMM Model file.

    Parse an hmmdef file and return dictionaries of hmms, states and
    transition matrices
    """
    return htk_model.load_model(filename)


def save_model(model, filename):
    """Save an HTK HMM Model file.
    """
    htk_model.save_model(model, filename)


def main():
    """Test code called from commandline"""
    model = load_model('../data/hmmdefs')
    hmm = model.hmms['r-We']
    for state_name in hmm.state_names:
        print(state_name)
        state = model.states[state_name]
        print(state.means_)
    print(model)
    model2 = load_model('../data/prior.hmm1mixSI.rate32')
    print(model2)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""load_htk_model
Reads an HTK model file

Usage:
    load_htk_model <model_file>... [options]
    load_htk_model --help

Options:
    <model_file>                                    the files of the model (e.g. macros hmmdefs)
    -o <output_file>, --output_file <output_file>   the output file (default is stdout)
    -h, --help                                      print this help screen
"""

from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = 'rmarxer'

import docopt
import numpy as np
import json
import collections

from . import htk_model_parser


def _to_ordered_dict(ast):
    result = collections.OrderedDict()
    for k, v in ast.items():
        result[k] = v

    return result


class HtkModelSemantics(object):
    def __init__(self):
        pass

    def matrix(self, ast):
        #return [float(v) for v in ast.split(' ')]
        return np.fromstring(ast, sep=' ')

    def vector(self, ast):
        #return [float(v) for v in ast.split(' ')]
        return np.fromstring(ast, sep=' ')

    def short(self, ast):
        return int(ast)

    def float(self, ast):
        return float(ast)

    def transPdef(self, ast):
        d = _to_ordered_dict(ast)
        d['matrix'] = d['array'].reshape((ast['dim'], ast['dim']))
        d.pop('array')
        return d

    def _default(self, ast):
        if isinstance(ast, collections.Mapping):
            return _to_ordered_dict(ast)

        return ast

    def _unquote(self, txt):
        if txt.startswith('"') and txt.endswith('"'):
            return txt[1:-1]

        return txt

    def string(self, ast):
        return self._unquote(ast)

    def __repr__(self):
        return ''


class NumPyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()  # or map(int, obj)
        return json.JSONEncoder.default(self, obj)


def load_model(*args):
    """Load an HTK model from one ore more files.

    :param args: Filenames of the model (e.g. macros hmmdefs)
    :return: The model as an OrderedDict()
    """
    text = ''
    for fnm in args:
        text += open(fnm).read()
        text += '\n'

    parser = htk_model_parser.htk_modelParser()
    model = HtkModelSemantics()
    return parser.parse(text,
                        rule_name='model',
                        ignorecase=True,
                        semantics=model,
                        comments_re="\(\*.*?\*\)",
                        trace=False)


def _array_to_htk(arr):
    result = ''

    for row in arr:
        result += ' {}\n'.format(' '.join(['{:2.6e}'.format(value) for value in row]))

    return result


def _serialize_option(option):
    result = ''
    if option.get('hmm_set_id', None) is not None:
        result += '<HmmSetId> {}'.format(option['hmm_set_id'])

    if option.get('stream_info', None) is not None:
        result += '<StreamInfo> {}'.format(option['stream_info']['count'])

        if option['stream_info'].get('sizes', None) is not None:
            result += ' {}'.format(' '.join(['{:d}'.format(v) for v in option['stream_info']['sizes']]))

    if option.get('vector_size', None) is not None:
        result += '<VecSize> {:d}'.format(option['vector_size'])

    if option.get('input_transform', None) is not None:
        raise NotImplementedError('Serialization of {} '
                                  'is not implemented.'.format(option['input_transform']))

    if option.get('covariance_kind', None) is not None:
        result += '<{}>'.format(option['covariance_kind'])

    if option.get('duration_kind', None) is not None:
        result += '<{}>'.format(option['duration_kind'])

    if option.get('parameter_kind', None) is not None:
        result += '<{}{}>'.format(option['parameter_kind']['base'],
                                  ''.join(option['parameter_kind']['options']))

    result += '\n'
    return result


def _serialize_transp(definition):
    if isinstance(definition, basestring):
        return '~t "{}"\n'.format(definition)

    result = ''
    result += '<TransP> {}\n'.format(definition['dim'])
    result += '{}'.format(_array_to_htk(definition['matrix']))
    return result


def _serialize_variance(definition):
    if isinstance(definition, basestring):
        return '~v {}\n'.format(definition)

    result = ''
    result += '<Variance> {}\n'.format(definition['dim'])
    result += '{}'.format(_array_to_htk(definition['vector'][None]))
    return result


def _serialize_mean(definition):
    if isinstance(definition, basestring):
        return '~u "{}"\n'.format(definition)

    result = ''
    result += '<Mean> {}\n'.format(definition['dim'])
    result += '{}'.format(_array_to_htk(definition['vector'][None]))
    return result


def _serialize_duration(definition):
    if isinstance(definition, basestring):
        return '~d "{}"\n'.format(definition)

    result = ''
    result += '<Duration> {}\n'.format(definition['dim'])
    result += '{}'.format(_array_to_htk(definition['vector'][None]))
    return result


def _serialize_weights(definition):
    if isinstance(definition, basestring):
        return '~w "{}"\n'.format(definition)

    result = ''
    result += '<SWeights> {}\n'.format(definition['dim'])
    result += '{}'.format(_array_to_htk(definition['vector'][None]))
    return result


def _serialize_covariance(definition):
    result = ''
    if definition['variance'] is not None:
        result += _serialize_variance(definition['variance'])

    else:
        raise NotImplementedError('Cannot serialize {}'.format(definition))

    return result


def _serialize_mixpdf(definition):
    if isinstance(definition, basestring):
        return '~m "{}"\n'.format(definition)

    result = ''
    if definition.get('regression_class', None) is not None:
        result += '<RClass> {}\n'.format(definition['regression_class'])

    result += _serialize_mean(definition['mean'])
    result += _serialize_covariance(definition['covariance'])

    if definition.get('gconst', None) is not None:
        result += '<GConst> {:.6e}\n'.format(definition['gconst'])

    return result


def _serialize_mixture(definition):
    result = ''

    if definition.get('index', None) is not None:
        result += '<Mixture> {} {:.6e}\n'.format(definition['index'], definition['weight'])

    result += _serialize_mixpdf(definition['pdf'])
    return result


def _serialize_stream(definition):
    result = ''

    if definition.get('dim', None) is not None:
        result += '<Stream> {}\n'.format(definition['dim'])

    if definition.get('mixtures', None) is not None:
        for mixture in definition['mixtures']:
            result += _serialize_mixture(mixture)

    else:
        raise NotImplementedError('Cannot serialize {}'.format(definition))

    return result


def _serialize_stateinfo(definition):
    if isinstance(definition, basestring):
        return '~s "{}"\n'.format(definition)

    result = ''
    if definition.get('streams_mixcount', None):
        result += '<NumMixes> {}\n'.format(' '.join(['{}'.format(v) for v in definition['streams_mixcount']]))

    if definition.get('weights', None) is not None:
        result += _serialize_weights(definition['weights'])

    for stream in definition['streams']:
        result += _serialize_stream(stream)

    if definition.get('duration', None) is not None:
        result += _serialize_duration(definition['duration'])

    return result


def _serialize_state(definition):
    result = ''

    result += '<State> {}\n'.format(definition['index'])
    result += _serialize_stateinfo(definition['state'])

    return result


def _serialize_hmm(definition):
    result = ''

    result += '<BeginHMM>\n'
    if definition.get('options', None):
        for option in definition['options']:
            result += _serialize_option(option)

    result += '<NumStates> {}\n'.format(definition['state_count'])

    for state in definition['states']:
        result += _serialize_state(state)

    if definition.get('regression_tree', None) is not None:
        raise NotImplementedError('Cannot serialize {}'.format(definition['regression_tree']))

    result += _serialize_transp(definition['transition'])

    if definition.get('duration', None) is not None:
        result += _serialize_duration(definition['duration'])

    result += '<EndHMM>\n'

    return result


def save_model(model, filename):
    """Save the model into a file.

    :param model: HTK model to be saved
    :param filename: File where to save the model
    """
    with open(filename, 'w') as f:
        f.write(serialize_model(model))


def serialize_model(model):
    """Serialize the HTK model into a file.

    :param model: Model to be serialized
    """

    result = ''

    # First serialize the macros
    for macro in model['macros']:
        if macro.get('options', None):
            result += '~o '
            for option in macro['options']['definition']:
                result += _serialize_option(option)

        elif macro.get('transition', None):
            result += '~t "{}"\n'.format(macro['transition']['name'])
            result += _serialize_transp(macro['transition']['definition'])

        elif macro.get('variance', None):
            result += '~v "{}"\n'.format(macro['variance']['name'])
            result += _serialize_variance(macro['variance']['definition'])

        elif macro.get('state', None):
            result += '~s "{}"\n'.format(macro['state']['name'])
            result += _serialize_stateinfo(macro['state']['definition'])

        elif macro.get('mean', None):
            result += '~u "{}"\n'.format(macro['mean']['name'])
            result += _serialize_mean(macro['mean']['definition'])

        elif macro.get('duration', None):
            result += '~d "{}"\n'.format(macro['duration']['name'])
            result += _serialize_duration(macro['duration']['definition'])

        else:
            raise NotImplementedError('Cannot serialize {}'.format(macro))

    for hmm in model['hmms']:
        if hmm.get('name', None) is not None:
            result += '~h "{}"\n'.format(hmm['name'])

        result += _serialize_hmm(hmm['definition'])

    return result


def model_to_json(model, **kwargs):
    """Convert a model to a JSON string.

    :param model: The OrderedDict representing the model
    :param kwargs:  Keyword arguments to be forwarded to the json.dumps method
    :return: The JSON string
    """
    return json.dumps(model, cls=NumPyArrayEncoder, **kwargs)


def main():
    arguments = docopt.docopt(__doc__)
    model = load_model(*arguments['<model_file>'])

    result = model_to_json(model, indent=2)

    if arguments['--output_file'] is not None:
        with open(arguments['--output_file'], 'w') as f:
            f.write(result)
    else:
        print(result)

    return

if __name__ == '__main__':
    main()


# inspire's setup.py
from setuptools import setup, find_packages

setup(
    name="inspire",
    packages=find_packages(),
    version="1.0.8",
    description="Helper library to participate in the INSPIRE challenge",
    author="Ricard Marxer",
    author_email="r.marxer@sheffield.ac.uk",
    url="https://github.com/rikrd/inspire",
    keywords=["intelligibility", "speech", "evaluation"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Linguistic",
    ],
    long_description="""\
INSPIRE Challenge library
-------------------------------------

This library includes a set of functions to help in the participation of the INSPIRE challenge.

The functions included in this module include:
 - loading the dataset
 - loading the lexicon
 - creating a participation
 - submitting a participation

Additionally it contains functions to create, train and use speech recognition models using HTK.
""",
    requires=['numpy', 'docopt', 'grako', 'progressbar_ipython', 'requests'],
    install_requires=['numpy', 'docopt', 'grako', 'progressbar_ipython', 'requests']
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Script for creating an ASR model
import os
import itertools
import shlex
import sys
import collections
import logging
import random
import tempfile
import json
import shutil
import unicodedata
import codecs
import re
import fnmatch
import copy
import math
import numpy as np

from . import config
from . import htk
from . import htk_model_utils

UTF8_NORMALIZATION = 'NFD'


def shlex_split(txt):
    return map(lambda s: s.decode('utf-8'), shlex.split(txt.encode('utf-8')))


def htk_to_utf8normalize(filename):
    with codecs.open(filename, 'r', 'string_escape') as f:
        data = unicodedata.normalize(UTF8_NORMALIZATION, f.read().decode('utf8'))

    with codecs.open(filename, 'w', 'utf-8') as f:
        f.write(data)

    return


def recursive_glob(path, pattern):
    return [os.path.join(dirpath, f)
            for dirpath, dirnames, files in os.walk(path)
            for f in fnmatch.filter(files, pattern)]


def parse_wordlist(list_words_filename):
    with codecs.open(list_words_filename, 'r', 'utf-8') as list_words_file:
        return [unicodedata.normalize(UTF8_NORMALIZATION, l).strip().upper() for l in list_words_file.readlines()]


def parse_dictionary(dictionary_filename):
    with codecs.open(dictionary_filename, 'r', 'utf-8') as f:
        return parses_dictionary(f.read())


def parses_dictionary(dictionary_buffer):
    dictionary = collections.defaultdict(set)

    for row, line in enumerate(dictionary_buffer.splitlines()):
        line = line.strip()
        if line.startswith('#'):
            continue

        try:
            key, value = line.split(None, 1)

        except ValueError:
            raise ValueError(u'Error parsing line {}.  "{}"'.format(row, line))

        dictionary[unicodedata.normalize(UTF8_NORMALIZATION, unicode(key))].add(
            unicodedata.normalize(UTF8_NORMALIZATION, unicode(value)))

    return dictionary


def check_words_in_dictionary(list_recordings, dictionary_filename, proposed_dictionary_filename=None):
    dictionary = parse_dictionary(dictionary_filename)
    dictionary_words = set(dictionary.keys())

    missing_words = set()
    proposed_pronunciations = collections.defaultdict(set)
    for recording in list_recordings:
        words = [w.upper() for w in get_words_from_recording(recording)]

        # if all words are in the dictionary we continue
        missing_recording_words = set(words) - dictionary_words
        if not missing_recording_words:
            continue

        missing_words |= missing_recording_words

        # otherwise we find their transcription
        proposed_pronunciations_recording = find_pronunciations(recording, missing_recording_words, dictionary)

        for key, value in proposed_pronunciations_recording.items():
            proposed_pronunciations[key] |= value

    if proposed_dictionary_filename:
        with codecs.open(proposed_dictionary_filename, 'w', 'utf-8') as f:
            for word, pronunciations in sorted(proposed_pronunciations.items()):
                for pronunciation in pronunciations:
                    f.write('{} {}\n'.format(word, pronunciation))

    return proposed_pronunciations


def find_pronunciations(recording, missing_words):
    proposed_dictionary = collections.defaultdict(set)

    words, phones = get_timit_from_recording(recording)
    if words is None or phones is None:
        return proposed_dictionary

    for missing_word in missing_words:
        # find the intervals where the missing_word appears
        intervals = [(int(v[0]), int(v[1])) for v in words if v[2].upper() == missing_word]

        for interval in intervals:
            # find the phonemes in the given interval
            phonemes = [v[2] for v in phones if int(v[0]) >= interval[0] and int(v[1]) <= interval[1]]
            pronunciation = ' '.join(phonemes)
            if pronunciation.startswith('sil '):
                pronunciation = pronunciation[len('sil '):]

            if pronunciation.endswith(' sil'):
                pronunciation = pronunciation[:-len(' sil')]

            if pronunciation:
                proposed_dictionary[missing_word].add(pronunciation)

    return proposed_dictionary


def get_speaker_from_inspire(fnm):
    return os.path.splitext(os.path.basename(fnm))[0].rsplit('_s', 1)[-1]


def get_speaker_from_wsjcam(fnm):
    speaker_filename = os.path.splitext(fnm)[0][:-2] + '00.ifo'
    if not os.path.isfile(speaker_filename):
        return None

    return os.path.split(os.path.dirname(fnm))[-1]


def get_timit_from_wsjcam(fnm):
    transcription_filename = os.path.splitext(fnm)[0] + '.wrd'
    if not os.path.isfile(transcription_filename):
        return None, None

    words = [unicodedata.normalize(UTF8_NORMALIZATION, line.strip().split()).upper() for line in
             codecs.open(transcription_filename, 'r', 'utf-8')]

    transcription_filename = os.path.splitext(fnm)[0] + '.phn'
    if not os.path.isfile(transcription_filename):
        return None, None

    phones = [line.strip().split() for line in codecs.open(transcription_filename, 'r', 'utf-8')]

    return words, phones


def get_words_from_wsjcam(fnm):
    transcription_filename = os.path.splitext(fnm)[0] + '.wrd'
    if not os.path.isfile(transcription_filename):
        return None

    words = [unicodedata.normalize(UTF8_NORMALIZATION, line.strip().split()[-1]).upper() for line in
             codecs.open(transcription_filename, 'r', 'utf-8')]

    return words


def get_words_from_inspire(fnm):
    # Always normalize when dealing with a filename (Mac +HFS uses NFD while all others use NFC normalisation)
    # Dictionaries are considered NFC
    return [unicodedata.normalize(UTF8_NORMALIZATION,
                                  os.path.splitext(os.path.basename(fnm))[0].rsplit('_s', 1)[0]).upper()]


def get_speaker_from_recording(fnm):
    return get_speaker_from_wsjcam(fnm) or get_speaker_from_inspire(fnm)


def get_words_from_recording(fnm):
    return get_words_from_wsjcam(fnm) or get_words_from_inspire(fnm)


def get_timit_from_recording(fnm):
    return get_timit_from_wsjcam(fnm)


def create_recordings_dictionary(list_selected, pronunciation_dictionary_filename, out_dictionary_filename, htk_trace,
                                 additional_dictionary_filenames=[]):
    """Create a pronunciation dictionary specific to a list of recordings"""
    temp_words_fd, temp_words_file = tempfile.mkstemp()

    words = set()
    for recording in list_selected:
        words |= set([w.upper() for w in get_words_from_recording(recording)])

    with codecs.open(temp_words_file, 'w', 'utf-8') as f:
        f.writelines([u'{}\n'.format(w) for w in sorted(list(words))])

    prepare_dictionary(temp_words_file,
                       pronunciation_dictionary_filename,
                       out_dictionary_filename,
                       htk_trace,
                       global_script_filename=config.project_path('etc/global.ded'),
                       additional_dictionary_filenames=additional_dictionary_filenames)

    os.remove(temp_words_file)

    return


def load_train_arguments(train_arguments_filename, args):
    # Set the non-supplied arguments using the training arguments
    with codecs.open(train_arguments_filename, 'r', 'utf-8') as train_arguments_file:
        train_args = json.load(train_arguments_file)

    if args.list_words_filename is None:
        args.list_words_filename = train_args['list_words_filename']

    if args.pronunciation_dictionary_filename is None:
        args.pronunciation_dictionary_filename = train_args['pronunciation_dictionary_filename']

    if args.feature_extractor_command is None:
        args.feature_extractor_command = train_args['feature_extractor_command']

    if args.feature_extractor_parameters is None:
        args.feature_extractor_parameters = train_args['feature_extractor_parameters']

    if not args.audio_dir:
        args.audio_dir = train_args['audio_dir']

    return args


def utf8_normalize(input_filename, comment_char='#', to_upper=False):
    """Normalize UTF-8 characters of a file
    """
    # Prepare the input dictionary file in UTF-8 and NFC
    temp_dict_fd, output_filename = tempfile.mkstemp()

    logging.debug('to_nfc from file {} to file {}'.format(input_filename, output_filename))

    with codecs.open(output_filename, 'w', 'utf-8') as f:
        with codecs.open(input_filename, 'r', 'utf-8') as input_f:
            lines = sorted([unicodedata.normalize(UTF8_NORMALIZATION, l)
                            for l in input_f.readlines()
                            if not l.strip().startswith(comment_char)])
            if to_upper:
                lines = [l.upper() for l in lines]

            f.writelines(lines)

    return output_filename


def merge_dictionaries(output_dictionary_filename, input_dictionary_filenames, htk_trace=0, global_script_filename=None,
                       add_sent_silence=True):
    if add_sent_silence:
        input_dictionary_filenames.append(config.project_path('etc', 'startendsilence.dict'))

    input_dictionary_in_nfc_filenames = []
    for filename in input_dictionary_filenames:
        output_filename = utf8_normalize(filename, to_upper=False)
        input_dictionary_in_nfc_filenames.append(output_filename)

    global_script = ""
    if global_script_filename:
        global_script = u"-g {}".format(global_script_filename)

    log_filename = config.temp_path('hdman.log')
    config.htk_command("HDMan -A -l {log_filename} -T {htk_trace} "
                       "{global_script} {output_dictionary_filename} "
                       "{dictionary_filenames}".format(htk_trace=htk_trace,
                                                       global_script=global_script,
                                                       dictionary_filenames=' '.join([filename
                                                                                      for filename
                                                                                      in
                                                                                      input_dictionary_in_nfc_filenames
                                                                                      if
                                                                                      os.path.getsize(filename) > 0]),
                                                       output_dictionary_filename=output_dictionary_filename,
                                                       log_filename=log_filename))

    os.remove(log_filename)
    for filename in input_dictionary_in_nfc_filenames:
        os.remove(filename)

    htk_to_utf8normalize(output_dictionary_filename)

    return


def prepare_dictionary(words_filename, dictionary_filename, output_dictionary_filename, htk_trace,
                       global_script_filename=None, add_sent_silence=True, additional_dictionary_filenames=[]):
    startendsilence_dictionary_filename = config.project_path('etc', 'startendsilence.dict')
    global_script = ""
    if global_script_filename:
        global_script = u"-g {}".format(global_script_filename)

    temp_dict_filename = utf8_normalize(dictionary_filename, to_upper=False)
    temp_words_filename = utf8_normalize(words_filename, to_upper=True)

    if add_sent_silence:
        with codecs.open(temp_words_filename, 'a', 'utf-8') as f:
            startendsilence = []
            with codecs.open(startendsilence_dictionary_filename, 'r', 'utf-8') as startendsilence_file:
                for line in startendsilence_file:
                    if line.strip():
                        startendsilence.append(line.strip().split(None, 1)[0])

            f.writelines([u'{}\n'.format(token) for token in startendsilence])

    log_filename = config.temp_path('hdman.log')
    config.htk_command("HDMan -A -l {log_filename} -T {htk_trace} "
                       "-w {words_filename} {global_script} "
                       "{output_dictionary_filename} {dictionary_filename} "
                       "{startendsilence_dictionary_filename} "
                       "{additional_dictionary_filenames}".format(htk_trace=htk_trace,
                                                                  words_filename=temp_words_filename,
                                                                  global_script=global_script,
                                                                  dictionary_filename=temp_dict_filename,
                                                                  startendsilence_dictionary_filename=startendsilence_dictionary_filename,
                                                                  output_dictionary_filename=output_dictionary_filename,
                                                                  log_filename=log_filename,
                                                                  additional_dictionary_filenames=' '.join(
                                                                      additional_dictionary_filenames)))

    # Parse the log file to look for missing words
    with codecs.open(log_filename, 'r', 'utf-8') as f:
        lines = f.read()

    missing_words = re.search(r'Missing Words\n-------------\n((.+\s)+)', lines, re.MULTILINE)
    if missing_words:
        words = missing_words.groups()[0].strip().split()
        raise ValueError(u"The following words could not be found by "
                         u"HTK in the pronunciation dictionary: \n{}".format(u'\n'.join(words)))

    os.remove(log_filename)
    os.remove(temp_words_filename)
    os.remove(temp_dict_filename)

    return


def parse_instance_count(instance_count, speaker_total_count):
    """This parses the instance count dictionary
    (that may contain floats from 0.0 to 1.0 representing a percentage)
    and converts it to actual instance count.
    """

    # Use all the instances of a speaker unless specified
    result = copy.copy(speaker_total_count)
    for speaker_id, count in instance_count.items():
        speaker_id = str(speaker_id)
        speaker_total = speaker_total_count.get(speaker_id, 0)

        if type(count) == float and 0.0 <= count <= 1.0:
            result[speaker_id] = int(speaker_total * count)

        else:
            result[speaker_id] = int(count)

    return result


def glob_recordings(recording_dirs, recording_patterns):
    for recording_dir in recording_dirs:
        if not os.path.isdir(recording_dir):
            logging.error("Recording directory {} supplied by user does not exist.".format(recording_dir))
            sys.exit(1)

    # Normalize to NFC (because Apple decided to be "original" and use NFD for their filesystem)
    recordings = []
    for recording_pattern in recording_patterns:
        for recording_dir in recording_dirs:
            recordings += [unicodedata.normalize(UTF8_NORMALIZATION, f) for f in
                           recursive_glob(unicode(recording_dir), recording_pattern)]

    return list(set(recordings))


def select_recording_set(recording_dirs, speaker_recording_instance_count, recordingset_filename, possible_words=None,
                         unaccepted_recordings=[]):
    unaccepted_recordings = set(unaccepted_recordings)

    recordings = recordings_possible = glob_recordings(recording_dirs, ['*.wav'])

    # Check if the words present in the recordings are all in the list of possible words if supplied
    if possible_words is not None:
        possible_words = set(possible_words)

        def all_words_possible(recording):
            return set([w.upper() for w in get_words_from_recording(recording)]) <= possible_words

        recordings_possible = filter(all_words_possible, recordings_possible)
        recordings_rejected = set(recordings) - set(recordings_possible)
        if len(recordings_rejected) > 0:
            logging.warning(u"Some recordings had to be rejected because "
                            "some of their words could not be found "
                            "in the possible words list: \n{}".format(u'\n'.join(recordings_rejected)))

    # Convert the speaker IDs to strings (to match the type of the IDs parsed from the recording filename)
    speaker_total_count = dict([(k, len(list(v)))
                                for k, v
                                in itertools.groupby(sorted(recordings_possible,
                                                            key=get_speaker_from_recording),
                                                     key=get_speaker_from_recording)])

    logging.debug('Speaker total count: {}'.format(speaker_total_count))
    speaker_recording_instance_count = parse_instance_count(speaker_recording_instance_count, speaker_total_count)

    logging.debug('Speaker recording instance count: {}'.format(speaker_recording_instance_count))

    recordings = filter(lambda x: x not in unaccepted_recordings, recordings_possible)

    # Check that there are no two recordings with the same basename (since this is used as ID in MLFs)
    recordings_ids_duplicates = {}
    for k, v in itertools.groupby(sorted(recordings, key=os.path.basename), key=os.path.basename):
        values = list(v)
        if len(values) > 1:
            recordings_ids_duplicates[k] = values

    if len(recordings_ids_duplicates) > 0:
        raise ValueError(
            "The recording filename basename must be unique since "
            "this is used as ID in the MLFs of HTK. Duplicated: {}".format(
                recordings_ids_duplicates))

    groups = itertools.groupby(sorted(recordings,
                                      key=get_speaker_from_recording),
                               key=get_speaker_from_recording)

    speaker_groups = [(k, list(v)) for k, v in groups]

    list_recording = []

    num_recording_instances = collections.defaultdict(lambda: sys.maxint)
    if speaker_recording_instance_count is not None:
        num_recording_instances = collections.defaultdict(lambda: 0)
        num_recording_instances.update(speaker_recording_instance_count)

    for i, (speaker_id, speaker_recordings) in enumerate(speaker_groups):
        num_instances = num_recording_instances[speaker_id]
        num_available = len(speaker_recordings)

        if num_instances > num_available:
            logging.warning(
                "Requested {} instances of speaker {} but only {} recordings are available.".format(num_instances,
                                                                                                    speaker_id,
                                                                                                    num_available))

        recording_instances = random.sample(speaker_recordings, min(num_instances, num_available))
        list_recording += recording_instances

    if len(list_recording) == 0:
        raise ValueError("No recordings could be selected.  Consider allowing the selection "
                         "of more recordings by removing filters --no-filter-training-recordings "
                         "or --no-filter-adaptation-recordings.")

    with codecs.open(recordingset_filename, 'w', 'utf-8') as f:
        f.write(u'\n'.join(list_recording))

    return list_recording


def create_feature_files(list_recordings, feature_directory, inputoutput_feature_file, output_feature_file,
                         feature_extractor, feature_extractor_parameters):
    list_inputoutput_feature = [(u'{}'.format(wavFile),
                                 u'{}.mfc'.format(os.path.join(feature_directory,
                                                               os.path.splitext(os.path.basename(wavFile))[0])))
                                for wavFile in list_recordings]

    # Create the feature space-separated input/output file
    with codecs.open(inputoutput_feature_file, 'w', 'utf-8') as f:
        f.writelines([u'"{}" "{}"\n'.format(wavFile, featureFile) for wavFile, featureFile in list_inputoutput_feature])

    # Create the file that contains the list of output feature files
    with codecs.open(output_feature_file, 'w', 'utf-8') as f:
        f.writelines([u'"{}"\n'.format(featureFile, ) for wavFile, featureFile in list_inputoutput_feature])

    cmd = u"{} {} {}".format(feature_extractor, feature_extractor_parameters, inputoutput_feature_file)
    logging.info(cmd)
    config.sh_command(cmd)

    return


def force_align(model_directory,
                inputoutput_feature_filename,
                symbollist_filename,
                word_mlf_filename,
                dictionary_filename,
                aligned_filename,
                output_words=False,
                output_likelihoods=False,
                output_times=False,
                level='word',
                boundary_silence=True,
                htk_trace=0):
    _aligned_file = aligned_filename
    if aligned_filename == word_mlf_filename:
        _aligned_file = config.temp_path('word_mlf_aligned.mlf')

    temp_dictionary_filename = utf8_normalize(dictionary_filename)

    output_arg = '{}{}{}'.format('' if output_likelihoods else 'S',
                                 '' if output_words else 'W',
                                 '' if output_times else 'T')

    levels = {'word': '',
              'state': '-f'}

    if level not in levels:
        raise ValueError('Only the following alignment levels are supported {}'.format(levels.keys()))

    level_arg = '{}'.format(levels[level])

    boundary_silence_str = '-b silence' if boundary_silence else ''

    config.htk_command("HVite -o '{}' {} -a -A -D -T {}"
                       " {} -S {}"
                       " -H {} -H {} -I {} -y lab"
                       " -i {} {} {}".format(output_arg,
                                             level_arg,
                                             htk_trace,
                                             boundary_silence_str,
                                             inputoutput_feature_filename,
                                             os.path.join(model_directory, 'macros'),
                                             os.path.join(model_directory, 'hmmdefs'),
                                             word_mlf_filename,
                                             _aligned_file,
                                             temp_dictionary_filename,
                                             symbollist_filename))
    os.remove(temp_dictionary_filename)

    if aligned_filename == word_mlf_filename:
        shutil.copyfile(_aligned_file, aligned_filename)
        os.remove(_aligned_file)

    return


def create_prototype_model(output_feature_filename,
                           output_prototype_filename,
                           state_stay_probabilities=[0.5]):
    with codecs.open(output_feature_filename, 'r', 'utf-8') as output_feature_file:
        example_feature_filename = shlex_split(output_feature_file.readline())[0]

    meta = htk.load_parameter(example_feature_filename, only_header=True)
    sample_dimension = int(meta.samp_size / 4)
    parameter_kind_base = meta.parm_kind_base
    parameter_kind_options = meta.parm_kind_opts

    proto_model = htk_model_utils.create_prototype(sample_dimension,
                                                   parameter_kind_base=parameter_kind_base,
                                                   parameter_kind_options=parameter_kind_options,
                                                   state_stay_probabilities=state_stay_probabilities)

    htk.save_model(proto_model, output_prototype_filename)


def create_flat_start_model(feature_filename,
                            state_stay_probabilities,
                            symbol_list,
                            output_model_directory,
                            output_prototype_filename,
                            htk_trace):
    """
    Creates a flat start model by using HCompV to compute the global mean and variance.
    Then uses these global mean and variance to create an N-state model for each symbol in the given list.

    :param feature_filename: The filename containing the audio and feature file pairs
    :param output_model_directory: The directory where to write the created model
    :param output_prototype_filename: The prototype model filename
    :param htk_trace: Trace level for HTK
    :rtype : None
    """
    # Create a prototype model
    create_prototype_model(feature_filename,
                           output_prototype_filename,
                           state_stay_probabilities=state_stay_probabilities)

    # Compute the global mean and variance
    config.htk_command("HCompV -A -D -T {} -f 0.01 "
                       "-S {} -m -o {} -M {} {}".format(htk_trace,
                                                        feature_filename,
                                                        'proto',
                                                        output_model_directory,
                                                        output_prototype_filename))

    # Create an hmmdefs using the global mean and variance for all states and symbols
    # Duplicate the model 'proto' -> symbol_list
    proto_model_filename = config.path(output_model_directory, 'proto')
    model = htk.load_model(proto_model_filename)
    model = htk_model_utils.map_hmms(model, {'proto': symbol_list})

    # vFloors -> macros
    vfloors_filename = config.path(output_model_directory, 'vFloors')
    variance_model = htk.load_model(vfloors_filename)

    model['macros'] += variance_model['macros']

    macros, hmmdefs = htk_model_utils.split_model(model)

    htk.save_model(macros, config.path(output_model_directory, 'macros'))
    htk.save_model(hmmdefs, config.path(output_model_directory, 'hmmdefs'))


def htk_mlf_symbol_replace(mlf,
                           replace_method=lambda x: x,
                           compress=True):
    new_mlf = {}
    for recording, label_seqs in mlf.items():
        new_label_seqs = []
        for label_seq in label_seqs:
            new_label_seq = []
            for label in label_seq:
                new_symbol = replace_method(label.symbol)
                label = label._replace(symbol=new_symbol)
                if compress \
                        and len(new_label_seq) \
                        and label.symbol == new_label_seq[-1].symbol:
                    # If we want to compress
                    # the new label sequence is not empty
                    # the new symbol is equal to the previous one
                    # we update the previous end time
                    new_label_seq[-1] = new_label_seq[-1]._replace(end=label.end)
                else:
                    # otherwise we append the new label
                    new_label_seq.append(label)

            new_label_seqs.append(new_label_seq)

        new_mlf[recording] = new_label_seqs

    return new_mlf


def train_speech_prior(model_directory,
                       component_count,
                       output_directory,
                       pronunciation_dictionary_filename,
                       list_recordings,
                       feature_filename,
                       speech_prior_mlf_filename,
                       htk_trace):
    models_directory = config.path('tmp/speech_prior', create=True)

    # Force align using the MLF triphone+sp files
    symbollist_filename = config.path(model_directory, 'symbollist')

    aligned_mlf_filename = config.path(models_directory, 'speech_prior_phone_aligned.mlf')
    reference_filename = config.path(models_directory, 'speech_prior_word.mlf')
    create_recordings_mlf(list_recordings,
                          pronunciation_dictionary_filename=pronunciation_dictionary_filename,
                          mlf_word_file=reference_filename,
                          htk_trace=htk_trace)

    logging.debug('Create a state alignment with the model for the silence/speech labels')
    force_align(model_directory,
                feature_filename,
                symbollist_filename,
                reference_filename,
                pronunciation_dictionary_filename,
                aligned_mlf_filename,
                output_times=True,
                output_likelihoods=False,
                output_words=False,
                level='state',
                htk_trace=htk_trace)

    # Modify the MLF files to only contain speech and sp symbols
    recognised_mlf = htk.load_mlf(aligned_mlf_filename, utf8_normalization=UTF8_NORMALIZATION)

    def only_silence_speech(x):
        if x.startswith('sil[') or x.startswith('sp['):
            return 'sil'
        else:
            return 'speech'

    replaced_mlf = htk_mlf_symbol_replace(recognised_mlf,
                                          replace_method=only_silence_speech,
                                          compress=True)

    htk.save_mlf(replaced_mlf, speech_prior_mlf_filename)

    # Initialize a single-state HMM model for speech
    logging.debug('Create flat start prior model')
    prior_prototype_filename = config.path(models_directory, 'proto.1state.mfc')
    model_directory = config.path(models_directory, 'prior.0', create=True)
    state_stay_probabilities = [0.5]
    symbol_list = ['speech', 'sil']
    create_flat_start_model(feature_filename,
                            state_stay_probabilities,
                            symbol_list,
                            model_directory,
                            prior_prototype_filename,
                            htk_trace)

    symbols_filename = config.path(models_directory, 'prior_symbols')
    with codecs.open(symbols_filename, 'w', 'utf-8') as f:
        f.write('\n'.join(symbol_list) + '\n')

    # Add GMM components and iterate
    logging.debug('Train prior model')
    # Reestimate parameters for monophone models (5 iterations)
    iterations = [1, 2, 3, 4, 5]
    for iteration in iterations:
        previous_model_directory = model_directory
        model_directory = config.path(models_directory, 'prior.{:03d}mix.{}'.format(1, iteration), create=True)

        config.htk_command("HERest -A -D -T {} -S {} -d {} "
                           "-H {} -H {} -I {} -M {} {}".format(htk_trace,
                                                               feature_filename,
                                                               previous_model_directory,
                                                               os.path.join(
                                                                   previous_model_directory,
                                                                   'macros'),
                                                               os.path.join(
                                                                   previous_model_directory,
                                                                   'hmmdefs'),
                                                               speech_prior_mlf_filename,
                                                               model_directory,
                                                               symbols_filename))

        # Copy the necessary to the model directory
        shutil.copy(symbols_filename, os.path.join(model_directory, 'symbollist'))

    # Increase the number of state mixtures
    mixtures = np.asarray(np.round(np.exp(np.linspace(np.log(2), np.log(component_count), 5))), dtype=np.int)
    for mixture in mixtures:
        previous_model_directory = model_directory
        model_directory = config.path(models_directory, 'prior.{:03d}mix.0'.format(mixture), create=True)

        script_filename = config.path(models_directory, 'mix{}.hed'.format(mixture))
        with codecs.open(script_filename, 'w', 'utf-8') as f:
            f.write('MU {} {{*.state[2-20].mix}}\n'.format(mixture))

        config.htk_command('HHEd -A -D -T {} -d {} '
                           '-H {} -H {} -M {} {} {}'.format(htk_trace,
                                                            previous_model_directory,
                                                            os.path.join(
                                                                previous_model_directory,
                                                                'macros'),
                                                            os.path.join(
                                                                previous_model_directory,
                                                                'hmmdefs'),
                                                            model_directory,
                                                            script_filename,
                                                            symbols_filename))

        # Copy the necessary to the model directory
        shutil.copy(symbols_filename, os.path.join(model_directory, 'symbollist'))

        iterations = [1, 2, 3, 4]
        for iteration in iterations:
            previous_model_directory = model_directory
            model_directory = config.path(models_directory, 'prior.{:03d}mix.{}'.format(mixture, iteration), create=True)

            config.htk_command("HERest -A -D -T {} -S {} -d {} "
                               "-H {} -H {} -I {} -M {} {}".format(htk_trace,
                                                                   feature_filename,
                                                                   previous_model_directory,
                                                                   os.path.join(
                                                                       previous_model_directory,
                                                                       'macros'),
                                                                   os.path.join(
                                                                       previous_model_directory,
                                                                       'hmmdefs'),
                                                                   speech_prior_mlf_filename,
                                                                   model_directory,
                                                                   symbols_filename))

            # Copy the necessary to the model directory
            shutil.copy(symbols_filename, os.path.join(model_directory, 'symbollist'))

    # Copy the final model files to the output directory
    shutil.copy(os.path.join(model_directory, 'macros'), os.path.join(output_directory, 'macros'))
    shutil.copy(os.path.join(model_directory, 'hmmdefs'), os.path.join(output_directory, 'hmmdefs'))
    shutil.copy(os.path.join(model_directory, 'symbollist'), os.path.join(output_directory, 'symbollist'))


def create_recordings_mlf(list_recordings,
                          pronunciation_dictionary_filename=None,
                          force_align_model_directory=None,
                          force_align_output_feature_filename=None,
                          force_align_symbollist_filename=None,
                          phones_filename=None,
                          phonessp_filename=None,
                          mlf_word_file=None,
                          mlf_phone_file=None,
                          mlf_phonesp_file=None,
                          mlf_tri_file=None,
                          mlf_trisp_file=None,
                          phone_script_filename=None,
                          phonesp_script_filename=None,
                          triphones_filename=None,
                          triphonessp_filename=None,
                          htk_trace=0):
    temp_files = []

    def create_and_track(x):
        f = config.temp_path(x)
        temp_files.append(f)
        return f

    _mlf_word_file = mlf_word_file or create_and_track('word.mlf')

    _mlf_phone_file = mlf_phone_file or create_and_track('phone.mlf')
    _mlf_phonesp_file = mlf_phonesp_file or create_and_track('phonesp.mlf')
    _phones_filename = phones_filename or create_and_track('phones')
    _phonessp_filename = phonessp_filename or create_and_track('phonessp')

    _mlf_tri_file = mlf_tri_file or create_and_track('tri.mlf')
    _mlf_trisp_file = mlf_trisp_file or create_and_track('trisp.mlf')
    _triphones_filename = triphones_filename or create_and_track('triphones')
    _triphonessp_filename = triphonessp_filename or create_and_track('triphonessp')

    phone_script_filename = phone_script_filename or config.project_path('etc', 'mkphones1.led')
    phonesp_script_filename = phonesp_script_filename or config.project_path('etc', 'mkphones1sp.led')

    compute_triphone = any([triphonessp_filename, triphones_filename, mlf_tri_file, mlf_trisp_file])
    compute_phone = any([compute_triphone, phonessp_filename, phones_filename, mlf_phone_file, mlf_phonesp_file])
    compute_word = any([compute_phone, mlf_word_file])

    if compute_word:
        logging.info("building word level MLFs")
        with codecs.open(_mlf_word_file, 'w', 'utf-8') as f:
            f.write("#!MLF!#\n")
            for recording in list_recordings:
                words = get_words_from_recording(recording)

                f.write(u'"*/{}.lab"\n'.format(os.path.splitext(os.path.basename(recording))[0]))
                f.writelines([u'{}\n'.format(word.upper()) for word in words])
                f.write(u'.\n\n')

                # Perform a force alignment using the passed model filename
    if force_align_model_directory \
            or force_align_output_feature_filename \
            or force_align_symbollist_filename:
        if not force_align_model_directory \
                or not force_align_output_feature_filename \
                or not force_align_symbollist_filename:
            raise ValueError("The model directory, the feature file and the symbol list"
                             " (wordlist, phonelist or tiedlist) are required when aligning.")

        force_align(force_align_model_directory,
                    force_align_output_feature_filename,
                    force_align_symbollist_filename,
                    _mlf_word_file,
                    pronunciation_dictionary_filename,
                    _mlf_word_file,
                    htk_trace)

    if compute_phone:
        if pronunciation_dictionary_filename is None:
            raise ValueError("A pronunciation dictionary file is required to generate phone level MLF.")

        temp_pronunciation_dictionary_filename = utf8_normalize(pronunciation_dictionary_filename)
        temp_files.append(temp_pronunciation_dictionary_filename)

        logging.info("building phone level MLFs")
        config.htk_command(
            "HLEd -A -C {} -D -T {} -l '*' -d {} -n {} -i {} {} {}".format(config.project_path('etc', 'config_empty'),
                                                                           htk_trace,
                                                                           temp_pronunciation_dictionary_filename,
                                                                           _phones_filename,
                                                                           _mlf_phone_file,
                                                                           phone_script_filename,
                                                                           _mlf_word_file))

        logging.info("building phone level and pause MLFs")
        config.htk_command(
            "HLEd -A -C {} -D -T {} -l '*' -d {} -n {} -i {} {} {}".format(config.project_path('etc', 'config_empty'),
                                                                           htk_trace,
                                                                           temp_pronunciation_dictionary_filename,
                                                                           _phonessp_filename,
                                                                           _mlf_phonesp_file,
                                                                           phonesp_script_filename,
                                                                           _mlf_word_file))

    if compute_triphone:
        logging.info("building triphone level MLFs")
        config.htk_command(
            "HLEd -A -C {} -D -T {} -l '*' -n {} -i {} {} {}".format(config.project_path('etc', 'config_empty'),
                                                                     htk_trace,
                                                                     _triphones_filename,
                                                                     _mlf_tri_file,
                                                                     config.project_path('etc/mktri.led'),
                                                                     _mlf_phone_file))

        logging.info("building triphone level and pause MLFs")
        config.htk_command(
            "HLEd -A -C {} -D -T {} -l '*' -n {} -i {} {} {}".format(config.project_path('etc', 'config_empty'),
                                                                     htk_trace,
                                                                     _triphonessp_filename,
                                                                     _mlf_trisp_file,
                                                                     config.project_path('etc/mktri.led'),
                                                                     _mlf_phonesp_file))

    for temp_file in temp_files:
        if os.path.exists(temp_file):
            os.remove(temp_file)

    return


"""
def create_triphone_dictionary(dictionary_filename,
                               triphone_dictionary_filename,
                               output_triphone_sybmol_filename = None,

                               htk_trace = 0):
    config.htk_command("HDMan -A -D -T {} -b sp "
                       "-n {} -g {} -l {} {} {}".format(htk_trace,
                                                        triphonehtk_filename,
                                                        config.project_path('etc',
                                                                            'global.tri.ded'),
                                                        log_filename,
                                                        pronunciation_triphone_dictionary_filename,
                                                        pronunciation_dictionary_filename))
"""


def get_dictionary_homophones(dictionary, words=None):
    inverse_dictionary = collections.defaultdict(set)
    for word, pronunciations in dictionary.items():
        if words is not None and word not in words:
            continue

        for pronunciation in pronunciations:
            inverse_dictionary[pronunciation].add(word)

    # TODO: Take possible words from the reference and recognition files
    # possible_words = {}
    # homophones = [(list(words)) for pronunciation, words
    # in inverse_dictionary.items() if len(words & possible_words) > 1]
    homophones = [list(words) for pronunciation, words in inverse_dictionary.items() if len(words) > 1]
    return homophones


def evaluate_model(reference_filename,
                   symbollist_filename,
                   recognition_filename,
                   dictionary_filename='',
                   htk_trace=0):
    ignore_arguments = '-e "???" "SENT-START" -e "???" "SENT-END" -e "???" sil -e "???" sp'

    # Only take words in the reference and recognition for the homophones
    mlf_ref = htk.load_mlf(reference_filename, UTF8_NORMALIZATION)
    mlf_rec = htk.load_mlf(recognition_filename, UTF8_NORMALIZATION)

    all_words = set()

    for hypotheses in mlf_ref.values() + mlf_rec.values():
        for label_seq in hypotheses:
            for label in label_seq:
                all_words.add(label.symbol)

    # Set equivalents based on homonyms
    homophones_arguments = u''
    if os.path.isfile(dictionary_filename):
        dictionary = parse_dictionary(dictionary_filename)
        homophones = get_dictionary_homophones(dictionary, words=all_words)
        for word_list in homophones:
            for word in word_list[1:]:
                homophones_arguments += u' -e "{}" "{}"'.format(word_list[0], word)

    cmd = u'HResults -T {htk_trace} -A {ignore_arguments} ' \
          u'{homophones_arguments} -I {reference_filename} ' \
          u'{symbollist_filename} {recognition_filename}'.format(reference_filename=reference_filename,
                                                                 symbollist_filename=symbollist_filename,
                                                                 recognition_filename=recognition_filename,
                                                                 ignore_arguments=ignore_arguments,
                                                                 homophones_arguments=homophones_arguments,
                                                                 htk_trace=max(1, htk_trace))
    config.htk_command(cmd)


def recognise_model(feature_filename,
                    symbollist_filename,
                    model_directory,
                    recognition_filename,
                    pronunciation_dictionary_filename,
                    list_words_filename='',
                    cmllr_directory=None,
                    tokens_count=None,
                    hypotheses_count=1,
                    htk_trace=0):
    """
        Perform recognition using a model and assuming a single word language.

        If the list_words_filename is == '' then all the words in the dictionary are used as language words.
    """
    # Normalize UTF-8 to avoid Mac problems
    temp_dictionary_filename = utf8_normalize(pronunciation_dictionary_filename)

    # Create language word list
    if list_words_filename:
        list_words = parse_wordlist(list_words_filename)
    else:
        list_words = sorted(parse_dictionary(temp_dictionary_filename).keys())

    # Create language model
    temp_directory = config.project_path('tmp', create=True)
    grammar_filename = config.path(temp_directory, 'grammar_words')
    wdnet_filename = config.path(temp_directory, 'wdnet')
    logging.debug('Create language model')
    create_language_model(list_words,
                          grammar_filename,
                          wdnet_filename)

    # Handle the Adaptation parameters
    cmllr_arguments = ''
    if cmllr_directory:
        if not os.path.isdir(cmllr_directory):
            logging.error('CMLLR adapatation directory not found: {}'.format(cmllr_directory))

        cmllr_arguments = "-J {} mllr2 -h '*/*_s%.mfc' -k -J {}".format(
            os.path.abspath(config.path(cmllr_directory, 'xforms')),
            os.path.abspath(config.path(cmllr_directory, 'classes')))
    # Handle the N-Best parameters
    hypotheses_count = hypotheses_count or 1
    tokens_count = tokens_count or int(math.ceil(hypotheses_count / 5.0))
    if hypotheses_count == 1 and tokens_count == 1:
        nbest_arguments = ""
    else:
        nbest_arguments = "-n {tokens_count} {hypotheses_count} ".format(tokens_count=tokens_count,
                                                                         hypotheses_count=hypotheses_count)

    # Run the HTK command
    config.htk_command("HVite -A -l '*' -T {htk_trace} "
                       "-H {model_directory}/macros -H {model_directory}/hmmdefs "
                       "-i {recognition_filename} -S {feature_filename} "
                       "{cmllr_arguments} -w {wdnet_filename} "
                       "{nbest_arguments} "
                       "-p 0.0 -s 5.0 "
                       "{pronunciation_dictionary_filename} "
                       "{symbollist_filename}".format(htk_trace=htk_trace,
                                                      model_directory=model_directory,
                                                      recognition_filename=recognition_filename,
                                                      feature_filename=feature_filename,
                                                      symbollist_filename=symbollist_filename,
                                                      nbest_arguments=nbest_arguments,
                                                      pronunciation_dictionary_filename=temp_dictionary_filename,
                                                      wdnet_filename=wdnet_filename,
                                                      cmllr_arguments=cmllr_arguments))

    # Remove temporary files
    os.remove(temp_dictionary_filename)
    os.remove(wdnet_filename)
    os.remove(grammar_filename)


def create_language_model(list_words,
                          grammar_filename,
                          wdnet_filename):
    """Create a language model (wdnet) for HTK using a very simple single-word grammar"""

    # Create the temporary grammar file
    grammar = u''
    grammar += u'$possible_words = '
    grammar += u' | '.join([word for word in list_words])
    grammar += u';\n'

    grammar += u'( [SENT-START] ( $possible_words ) [SENT-END] )\n'

    with codecs.open(grammar_filename, 'w', 'utf-8') as f:
        f.write(grammar)

    # Use HParse to create the wdnet file
    config.htk_command('HParse -A {} {}'.format(grammar_filename, wdnet_filename))

    return
